/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */

#ifndef ALGO_QA_HISTO1D_H
#define ALGO_QA_HISTO1D_H 1

#include <cstdint>
#include <string>
#include <vector>


namespace cbm::algo
{

  /** @class Histo1D
   ** @author Volker Friese <v.friese@gsi.de>
   ** @since 15 June 2023
   **
   ** Lightweight one-dimensional histogram class
   **/
  class Histo1D {
  public:
    /** @brief No default constructor **/
    Histo1D() = delete;


    /** @brief Standard constructor
     ** @param numBins   Number of bins
     ** @param minValue  Lower edge of histogram range
     ** @param maxValue  Upper edge of histogram range
     ** @param name      Histogram name
     **/
    Histo1D(uint32_t numBins, double minValue, double maxValue, const std::string& name = "");


    /** @brief Destructor **/
    virtual ~Histo1D() = default;


    /** @brief Add an entry to the histogram
     ** @param value  Value to be histogrammed
     ** @param weight Weight for the entry
     **/
    void Add(double value, double weight = 1.);


    /** @brief Clear histogram contents **/
    void Clear();


    /** @brief Histogram content in a bin
     ** @param bin Bin index
     ** @return Histogram content. Zero if bin index is out of scope.
     **/
    double Content(uint32_t bin) const;


    /** @brief Number of entries **/
    double NumEntries() const { return fNumEntries; }


    /** @brief Upper edge **/
    double MaxValue() const { return fMaxValue; }


    /** @brief First moment of distribution **/
    double Mean() const;


    /** @brief Lower edge **/
    double MinValue() const { return fMinValue; }


    /** @brief Histogram name **/
    const std::string& Name() const { return fName; }


    /** @brief Number of bins **/
    uint32_t NumBins() const { return fNumBins; }

    /** @brief Add another histogram to an existing one
     ** @param other Histogram object to be added
     **/
    Histo1D& operator+=(const Histo1D& other);


    /** @brief Overflow **/
    double Overflow() const { return fOverflow; }


    /** @brief Second moment of distribution **/
    double Stddev() const;


    /** @brief Properties to string **/
    std::string ToString() const;


    /** @brief Underflow **/
    double UnderFlow() const { return fUnderflow; }


  private:
    /** Properties **/
    const std::string fName;
    const uint32_t fNumBins;
    const double fMinValue;
    const double fMaxValue;

    /** Content **/
    std::vector<double> fContent = {};
    double fUnderflow            = 0;
    double fOverflow             = 0;
    size_t fNumEntries           = 0;
  };


  /** @brief Adding two histograms
   ** @param h1,h2 Histograms to be added
   ** @return Sum histogram
   **/
  inline Histo1D operator+(Histo1D h1, const Histo1D& h2)
  {
    h1 += h2;
    return h1;
  }

} /* namespace cbm::algo */

#endif /* ALGO_QA_HISTO1D_H */
