/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/***************************************************************************************************
 * @file   CbmMvdTrackingInterface.h
 * @brief  Input data and parameters interface from MVD subsystem used in L1 tracker (declaration)
 * @since  31.05.2022
 * @author S.Zharko <s.zharko@gsi.de>
 ***************************************************************************************************/

#ifndef CbmMvdTrackingInterface_h
#define CbmMvdTrackingInterface_h 1

#include "CbmMvdDetectorId.h"                  // for CbmMvdDetectorId
#include "CbmMvdHit.h"                         // for CbmMvdHit
#include "CbmMvdStationPar.h"                  // for CbmMvdStationPar
#include "CbmPixelHit.h"                       // for CbmPixelHit
#include "CbmTrackingDetectorInterfaceBase.h"  // for CbmTrackingDetectorInt...

#include <FairTask.h>  // for InitStatus, FairTask

#include <Rtypes.h>  // for ClassDef
#include <TMath.h>   // for Pi

#include <algorithm>  // for min, max
#include <string>     // for string

class TBuffer;
class TClass;
class TMemberInspector;

/// Class CbmMvdTrackingInterface is a CbmL1 subtask, which provides necessary methods for L1 tracker
/// to access the geometry and dataflow settings.
///
class CbmMvdTrackingInterface : public FairTask, public CbmTrackingDetectorInterfaceBase, public CbmMvdDetectorId {
public:
  /// Default constructor
  CbmMvdTrackingInterface();

  /// Destructor
  ~CbmMvdTrackingInterface();

  /// FairTask: Init method
  InitStatus Init();

  /// FairTask: ReInit method
  InitStatus ReInit();

  /// Gets pointer to the instance of the CbmMvdTrackingInterface
  static CbmMvdTrackingInterface* Instance() { return fpInstance; }

  /// Gets name of this subsystem
  std::string GetDetectorName() const { return "MVD"; }

  /// Gets actual number of tracking stations, provided by the current geometry setup
  int GetNtrackingStations() const { return fMvdStationPar->GetStationCount(); }

  /// Gets reference z of the station
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Z position of the station [cm]
  double GetZref(int stationId) const { return fMvdStationPar->GetZPosition(stationId); }

  /// Gets min z of the station
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return min Z of the station [cm]
  double GetZmin(int stationId) const { return GetZref(stationId) - 0.5 * fMvdStationPar->GetZThickness(stationId); }

  /// Gets max z of the station
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return max Z of the station [cm]
  double GetZmax(int stationId) const { return GetZref(stationId) + 0.5 * fMvdStationPar->GetZThickness(stationId); }

  /// Gets a tracking station of a CbmPixelHit
  /// \param  hit  A pointer to CbmPixelHit
  /// \return Local index of the tracking station
  int GetTrackingStationIndex(const CbmPixelHit* hit) const
  {
    auto hitMvd = [&] {
      if constexpr (kUseDynamicCast) { return dynamic_cast<const CbmMvdHit*>(hit); }
      else {
        return static_cast<const CbmMvdHit*>(hit);
      }
    }();
    return hitMvd->GetStationNr();
  }

  /// Gets a tracking station by the address of element (detectorID in terms of MVD)
  /// \param  detectorId  Unique element address (detectorID in terms of MVD)
  /// \return Local index of the tracking station
  int GetTrackingStationIndex(int detectorId) const { return StationNr(detectorId); }

  /// Gets max size of a tracking station along the X-axis
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Size of station along the X-axis [cm]
  double GetXmax(int stationId) const { return fMvdStationPar->GetWidth(stationId); }

  /// Gets max size of a tracking station along the Y-axis
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Size of station along the Y-axis [cm]
  double GetYmax(int stationId) const { return fMvdStationPar->GetHeight(stationId); }

  /// Check if the detector provides time measurements
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Flag: true - station provides time measurements, false - station does not provide time measurements
  bool IsTimeInfoProvided(int /*stationId*/) const { return false; }

  /// Gets stereo angles of the two independent measured coordinates
  /// The tracking does not use this method. It is only used by the QA task.
  /// \param  address  detector unique identifier
  /// \return [phiU, phiV] - Stereo angles [rad]
  std::tuple<double, double> GetStereoAnglesSensor(int /*address*/) const { return std::tuple(0., TMath::Pi() / 2.); }

  /// FairTask: sets parameter containers up
  void SetParContainers();

  /// --- to be removed --- Gets the tracking station radiation length
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Radiation length [cm]
  double GetRadLength(int stationId) const
  {
    return fMvdStationPar->GetZThickness(stationId) / (10. * fMvdStationPar->GetZRadThickness(stationId));
  }

  /// --- to be removed --- Gets the tracking station thickness along the Z-axis
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Station thickness [cm]
  double GetSensorThickness(int stationId) const { return fMvdStationPar->GetZThickness(stationId); }

  /// Copy and move constructers and assign operators are prohibited
  CbmMvdTrackingInterface(const CbmMvdTrackingInterface&) = delete;
  CbmMvdTrackingInterface(CbmMvdTrackingInterface&&)      = delete;
  CbmMvdTrackingInterface& operator=(const CbmMvdTrackingInterface&) = delete;
  CbmMvdTrackingInterface& operator=(CbmMvdTrackingInterface&&) = delete;

private:
  static CbmMvdTrackingInterface* fpInstance;  ///< Instance of the class

  const CbmMvdStationPar* fMvdStationPar {nullptr};  ///< Pointer to the Mvd station parameters

  ClassDef(CbmMvdTrackingInterface, 0);
};

#endif  // CbmMvdTrackingInterface
