/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/***************************************************************************************************
 * @file   CbmTrdTrackingInterface.h
 * @brief  Input data and parameters interface from TRD subsystem used in L1 tracker (declaration)
 * @since  31.05.2022
 * @author S.Zharko <s.zharko@gsi.de>
 ***************************************************************************************************/

#ifndef CbmTrdTrackingInterface_h
#define CbmTrdTrackingInterface_h 1

#include "CbmPixelHit.h"
#include "CbmTrackingDetectorInterfaceBase.h"
#include "CbmTrdAddress.h"
#include "CbmTrdParModDigi.h"
#include "CbmTrdParSetDigi.h"

#include "FairTask.h"

#include "TMath.h"

#include <iostream>
#include <vector>

/// Class CbmTrdTrackingInterface is a CbmL1 subtask, which provides necessary methods for L1 tracker
/// to access the geometry and dataflow settings.
///
/// NOTE: For TRD one tracking station is a TRD module!
///
class CbmTrdTrackingInterface : public FairTask, public CbmTrackingDetectorInterfaceBase {
public:
  /// Default constructor
  CbmTrdTrackingInterface();

  /// Destructor
  ~CbmTrdTrackingInterface();

  /// FairTask: Init method
  InitStatus Init();

  /// FairTask: ReInit method
  InitStatus ReInit();

  /// Gets pointer to the instance of the CbmTrdTrackingInterface
  static CbmTrdTrackingInterface* Instance() { return fpInstance; }

  /// Gets name of this subsystem
  std::string GetDetectorName() const { return "TRD"; }

  /// Gets actual number of tracking stations, provided by the current geometry setup
  int GetNtrackingStations() const;

  /// Gets reference z of the station
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Z position of the station [cm]
  double GetZref(int stationId) const { return GetTrdModulePar(stationId)->GetZ(); }

  /// Gets min z of the station
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return min Z of the station [cm]
  double GetZmin(int stationId) const { return GetZref(stationId) - GetTrdModulePar(stationId)->GetSizeZ(); }

  /// Gets max z of the station
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return max Z of the station [cm]
  double GetZmax(int stationId) const { return GetZref(stationId) + GetTrdModulePar(stationId)->GetSizeZ(); }

  /// Gets a tracking station of a CbmPixelHit
  /// \param  hit  A pointer to CbmPixelHit
  /// \return Local index of the tracking station
  int GetTrackingStationIndex(const CbmPixelHit* hit) const { return GetTrackingStationIndex(hit->GetAddress()); }

  /// Gets a tracking station by the address
  /// \param  address  Unique element address
  /// \return Local index of the tracking station
  int GetTrackingStationIndex(int address) const { return CbmTrdAddress::GetLayerId(address); }

  /// Gets max size of a station along the X-axis
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Size of station along the X-axis [cm]
  double GetXmax(int stationId) const { return GetTrdModulePar(stationId)->GetSizeX(); }

  /// Gets max size of a station along the Y-axis
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Size of station along the Y-axis [cm]
  double GetYmax(int stationId) const { return GetTrdModulePar(stationId)->GetSizeY(); }

  /// Check if station provides time measurements
  /// \param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// \return Flag: true - station provides time measurements, false - station does not provide time measurements
  bool IsTimeInfoProvided(int /*stationId*/) const { return true; }

  /// Gets stereo angles of the two independent measured coordinates
  /// The tracking does not use this method. It is only used by the QA task.
  /// \param  address  detector unique identifier
  /// \return [phiU, phiV] - Stereo angles [rad]
  std::tuple<double, double> GetStereoAnglesSensor(int /*address*/) const { return std::tuple(0., TMath::Pi() / 2.); }

  /// FairTask: sets parameter containers up
  void SetParContainers();

  /// Copy and move constructers and assign operators are prohibited
  CbmTrdTrackingInterface(const CbmTrdTrackingInterface&) = delete;
  CbmTrdTrackingInterface(CbmTrdTrackingInterface&&)      = delete;
  CbmTrdTrackingInterface& operator=(const CbmTrdTrackingInterface&) = delete;
  CbmTrdTrackingInterface& operator=(CbmTrdTrackingInterface&&) = delete;

private:
  /// Gets pointer to the TRD module
  /// \param  moduleId  Id of the Trd module
  /// \return Pointer to the particular CbmTrdParModDigi object
  __attribute__((always_inline)) CbmTrdParModDigi* GetTrdModulePar(int moduleId) const
  {
    return static_cast<CbmTrdParModDigi*>(fTrdDigiPar->GetModulePar(fTrdDigiPar->GetModuleId(moduleId)));
  }

  static CbmTrdTrackingInterface* fpInstance;  ///< Instance of the class

  CbmTrdParSetDigi* fTrdDigiPar {nullptr};
  //CbmTrdParModDigi* fTrdModuleInfo {nullptr};

  ClassDef(CbmTrdTrackingInterface, 0);
};

#endif  // CbmTrdTrackingInterface
