/* Copyright (C) 2006-2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Ivan Kisel, Sergey Gorbunov [committer], Valentina Akishina, Igor Kulakov, Maksym Zyzak, Sergei Zharko */

/*
 *====================================================================
 *
 *  CBM Level 1 Reconstruction
 *
 *  Authors: I.Kisel,  S.Gorbunov
 *
 *  e-mail : ikisel@kip.uni-heidelberg.de
 *
 *====================================================================
 *
 *  L1 main program
 *
 *====================================================================
 */

#include "CbmL1.h"

#include "CbmMCDataManager.h"
#include "CbmMuchTrackingInterface.h"
#include "CbmMvdTrackingInterface.h"
#include "CbmSetup.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTofTrackingInterface.h"
#include "CbmTrdTrackingInterface.h"

#include <boost/filesystem.hpp>

#include "CaToolsMaterialHelper.h"
// TODO: include of CbmSetup.h creates problems on Mac
// #include "CbmSetup.h"
#include "CbmEvent.h"
#include "CbmMCDataObject.h"
#include "CbmStsFindTracks.h"
#include "CbmStsHit.h"
#include "CbmTrackingDetectorInterfaceInit.h"

#include "FairEventHeader.h"
#include "FairField.h"
#include "FairRunAna.h"

#include "TGeoArb8.h"
#include "TGeoBoolNode.h"
#include "TGeoCompositeShape.h"
#include "TGeoManager.h"
#include "TGeoNode.h"
#include "TMatrixD.h"
#include "TNtuple.h"
#include "TProfile2D.h"
#include "TROOT.h"
#include "TRandom3.h"
#include "TSystem.h"
#include "TVector3.h"
#include "TVectorD.h"
#include <TFile.h>

#include <chrono>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <list>

#include "L1Algo/L1Algo.h"
#include "L1Algo/L1Assert.h"
#include "L1Algo/L1Branch.h"
#include "L1Algo/L1Field.h"
#include "L1Algo/L1Hit.h"
#include "L1Event.h"

using std::cout;
using std::endl;
using std::ios;

ClassImp(CbmL1);

static L1Algo gAlgo _fvecalignment;  // TODO: Change coupling logic between L1Algo and CbmL1

CbmL1* CbmL1::fpInstance = 0;


// ---------------------------------------------------------------------------------------------------------------------
//
CbmL1::CbmL1() : CbmL1("L1") {}

// ---------------------------------------------------------------------------------------------------------------------
//
CbmL1::CbmL1(const char* name, Int_t verbose, Int_t performance) : FairTask(name, verbose), fPerformance(performance)
{
  if (!fpInstance) fpInstance = this;

  switch (fSTAPDataMode) {
    case 1: LOG(info) << "CbmL1: input data will be written for a standalone usage"; break;
    case 2: LOG(info) << "CbmL1: input data will be read from external files"; break;
    default: LOG(info) << "CbmL1: tracking will be run without external data R/W"; break;
  }

  fpIODataManager = std::make_shared<L1IODataManager>();

  this->DefineSTAPNames("");

  if (!CbmTrackingDetectorInterfaceInit::Instance()) {
    LOG(fatal) << "CbmL1: CbmTrackingDetectorInterfaceInit instance was not found. Please, add it as a task to your "
                  "reco macro right before the KF and L1 tasks:\n"
               << "\033[1;30mrun->AddTask(new CbmTrackingDetectorInterfaceInit());\033[0m";
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
CbmL1::~CbmL1()
{
  if (fpInstance == this) fpInstance = nullptr;
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::CheckDetectorPresence()
{
  fUseMVD  = fUseMVD && CbmSetup::Instance()->IsActive(ECbmModuleId::kMvd);
  fUseSTS  = fUseSTS && CbmSetup::Instance()->IsActive(ECbmModuleId::kSts);
  fUseMUCH = fUseMUCH && CbmSetup::Instance()->IsActive(ECbmModuleId::kMuch);
  fUseTRD  = fUseTRD && CbmSetup::Instance()->IsActive(ECbmModuleId::kTrd);
  fUseTOF  = fUseTOF && CbmSetup::Instance()->IsActive(ECbmModuleId::kTof);
  {
    // TODO: temporary code!!
    // for a moment, the MVD digitizer doesn't work in TB mode
    // check the presence of MVD hits to make sure the MVD is really active
    if (!FairRootManager::Instance()->GetObject("MvdHit")) { fUseMVD = false; }
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::DisableTrackingStation(L1DetectorID detID, int iSt)
{
  if (L1DetectorID::kEND != detID) { fvmDisabledStationIDs[detID].insert(iSt); }
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmL1::ReInit()
{
  SetParContainers();
  return Init();
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmL1::Init()
{
  if (fVerbose > 1) {
    char y[20] = " [0;33;44m";         // yellow
    char Y[20] = " [1;33;44m";         // yellow bold
    char W[20] = " [1;37;44m";         // white bold
    char o[20] = " [0m\n";             // color off
    Y[0] = y[0] = W[0] = o[0] = 0x1B;  // escape character

    cout << endl << endl;
    cout << "  " << W << "                                                                 " << o;
    cout << "  " << W << "  ===////======================================================  " << o;
    cout << "  " << W << "  =                                                           =  " << o;
    cout << "  " << W << "  =                   " << Y << "L1 on-line finder" << W << "                       =  " << o;
    cout << "  " << W << "  =                                                           =  " << o;
    cout << "  " << W << "  =  " << W << "Cellular Automaton 3.1 Vector" << y << " with " << W << "KF Quadro" << y
         << " technology" << W << "  =  " << o;
    cout << "  " << W << "  =                                                           =  " << o;
    cout << "  " << W << "  =  " << y << "Designed for CBM collaboration" << W << "                           =  " << o;
    cout << "  " << W << "  =  " << y << "All rights reserved" << W << "                                      =  " << o;
    cout << "  " << W << "  =                                                           =  " << o;
    cout << "  " << W << "  ========================================================////=  " << o;
    cout << "  " << W << "                                                                 " << o;
    cout << endl << endl;
  }

  if (fVerbose > 1) {
#ifdef _OPENMP
    LOG(info) << "L1 is running with OpenMP" << endl;
#else
    LOG(info) << "L1 is running without OpenMP" << endl;
#endif
  }

  fHistoDir = gROOT->mkdir("L1");
  fHistoDir->mkdir("Input");
  fHistoDir->mkdir("Fit");

  fTableDir = gROOT->mkdir("L1Tables");

  // turn on reconstruction in sub-detectors

  fUseMVD  = false;
  fUseSTS  = false;
  fUseMUCH = false;
  fUseTRD  = false;
  fUseTOF  = false;

  FairRootManager* fairManager = FairRootManager::Instance();

  if (L1Algo::TrackingMode::kSts == fTrackingMode) {
    fUseMVD  = 1;
    fUseSTS  = 1;
    fUseMUCH = 0;
    fUseTRD  = 0;
    fUseTOF  = 0;
    // check if MVD is switched off in the Sts task
    CbmStsFindTracks* findTask = L1_DYNAMIC_CAST<CbmStsFindTracks*>(FairRunAna::Instance()->GetTask("STSFindTracks"));
    if (findTask) fUseMVD = findTask->MvdUsage();
  }

  if (L1Algo::TrackingMode::kMcbm == fTrackingMode) {
    fUseMVD  = 1;
    fUseSTS  = 1;
    fUseMUCH = 1;
    fUseTRD  = 1;
    fUseTOF  = 1;
    fInitManager.DevSetIgnoreHitSearchAreas(true);  // uncomment for debug
  }

  if (L1Algo::TrackingMode::kGlobal == fTrackingMode) {
    //at the moment trd2d tracking only
    fUseMVD  = 1;
    fUseSTS  = 1;
    fUseMUCH = 1;
    fUseTRD  = 1;
    fUseTOF  = 0;

    fInitManager.DevSetUseOfOriginalField();
    fInitManager.DevSetIgnoreHitSearchAreas(true);
    fInitManager.SetMaxTripletPerDoublets(1000);
  }


  //fInitManager.DevSetIsMatchDoubletsViaMc(true);
  //fInitManager.DevSetIsMatchTripletsViaMc(true);
  //fInitManager.DevSetIsExtendTracksViaMc(true);
  //fInitManager.DevSetIsSuppressOverlapHitsViaMc(true);

  CheckDetectorPresence();

  fpStsPoints  = 0;
  fpMvdPoints  = 0;
  fpMuchPoints = 0;
  fpTrdPoints  = 0;
  fpTofPoints  = 0;
  fpMcTracks   = 0;

  fpMvdHitMatches  = 0;
  fpTrdHitMatches  = 0;
  fpMuchHitMatches = 0;
  fpTofHitMatches  = 0;

  fpStsClusters = 0;

  fvSelectedMcEvents.clear();

  fTimeSlice = (CbmTimeSlice*) fairManager->GetObject("TimeSlice.");
  if (!fTimeSlice) { LOG(fatal) << GetName() << ": No time slice branch in the tree!"; }

  fpStsClusters       = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("StsCluster"));
  fpStsHitMatches     = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("StsHitMatch"));
  fpStsClusterMatches = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("StsClusterMatch"));

  fpStsHits = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("StsHit"));

  if (!fUseMUCH) {
    fpMuchPixelHits   = nullptr;
    fpMuchDigis       = nullptr;  // NOTE: Used only with fPerformance = true
    fpMuchDigiMatches = nullptr;  // NOTE: Used only with fPerformance = true
    fpMuchClusters    = nullptr;  // NOTE: Used only with fPerformance = true
    fpMuchPoints      = nullptr;  // NOTE: Used only with fPerformance = true
    fpMuchHitMatches  = nullptr;  // NOTE: Used only with fPerformance = true
  }
  else {
    fpMuchPixelHits = (TClonesArray*) fairManager->GetObject("MuchPixelHit");
  }

  if (!fUseTRD) {
    fpTrdPoints     = nullptr;
    fpTrdHitMatches = nullptr;
    fpTrdHits       = nullptr;
  }
  else {
    fpTrdHits = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("TrdHit"));
  }

  if (!fUseTOF) {
    fpTofPoints     = nullptr;
    fpTofHitMatches = nullptr;
    fpTofHits       = nullptr;
  }
  else {
    fpTofHits = (TClonesArray*) fairManager->GetObject("TofHit");
  }

  if (fPerformance) {
    CbmMCDataManager* mcManager = (CbmMCDataManager*) fairManager->GetObject("MCDataManager");
    if (nullptr == mcManager) LOG(fatal) << GetName() << ": No CbmMCDataManager!";

    fpMcEventHeader = mcManager->GetObject("MCEventHeader.");

    fpMcTracks = mcManager->InitBranch("MCTrack");

    if (nullptr == fpMcTracks) LOG(fatal) << GetName() << ": No MCTrack data!";
    if (nullptr == fpMcEventHeader) LOG(fatal) << GetName() << ": No MC event header data!";

    fpMcEventList = (CbmMCEventList*) fairManager->GetObject("MCEventList.");
    if (nullptr == fpMcEventList) LOG(fatal) << GetName() << ": No MCEventList data!";

    if (fUseMVD) {
      fpMvdPoints      = mcManager->InitBranch("MvdPoint");
      fpMvdDigiMatches = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("MvdDigiMatch"));
      fpMvdHitMatches  = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("MvdHitMatch"));
      if (!fpMvdHitMatches) { LOG(error) << "No fpMvdHitMatches provided, performance is not done correctly"; }
    }

    if (fUseSTS) {
      fpStsPoints = mcManager->InitBranch("StsPoint");
      if (nullptr == fpStsPoints) LOG(fatal) << GetName() << ": No StsPoint data!";
    }

    fpTrdPoints = mcManager->InitBranch("TrdPoint");

    if (!fUseTRD) { fpTrdHitMatches = 0; }
    else {
      fpTrdHitMatches = (TClonesArray*) fairManager->GetObject("TrdHitMatch");
    }

    if (!fUseMUCH) {
      fpMuchPoints     = 0;
      fpMuchHitMatches = 0;
    }
    else {
      fpMuchDigis       = (TClonesArray*) fairManager->GetObject("MuchDigi");
      fpMuchDigiMatches = (TClonesArray*) fairManager->GetObject("MuchDigiMatch");
      fpMuchClusters    = (TClonesArray*) fairManager->GetObject("MuchCluster");
      fpMuchPoints      = mcManager->InitBranch("MuchPoint");
      fpMuchHitMatches  = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("MuchPixelHitMatch"));
    }

    fpTofPoints = mcManager->InitBranch("TofPoint");

    if (!fUseTOF) { fpTofHitMatches = 0; }
    else {
      fpTofHitMatches = static_cast<TClonesArray*>(fairManager->GetObject("TofHitMatch"));
    }
  }

  if (!fUseMVD) { fpMvdHits = 0; }
  else {
    fpMvdHits = L1_DYNAMIC_CAST<TClonesArray*>(fairManager->GetObject("MvdHit"));
  }

  // *****************************
  // **                         **
  // ** GEOMETRY INITIALIZATION **
  // **                         **
  // *****************************

  // Read parameters object from a binary file
  if (2 == fSTAPDataMode) {
    this->ReadSTAPParamObject();

    fNMvdStationsGeom  = fInitManager.GetNstationsGeometry(L1DetectorID::kMvd);
    fNStsStationsGeom  = fInitManager.GetNstationsGeometry(L1DetectorID::kSts);
    fNTrdStationsGeom  = fInitManager.GetNstationsGeometry(L1DetectorID::kTrd);
    fNMuchStationsGeom = fInitManager.GetNstationsGeometry(L1DetectorID::kMuch);
    fNTofStationsGeom  = fInitManager.GetNstationsGeometry(L1DetectorID::kTof);
    fNStationsGeom     = fInitManager.GetNstationsGeometry();
  }
  // Parameters initialization, based on the FairRunAna chain
  else {
    // ********************************************
    // ** Base configuration file initialization **
    // ********************************************
    {
      // The parameters module of CA algorithm (L1Parameters) contains two sets of data fields: the geometry configura-
      // tion and the parameter configuration. The configuration both for geometry and parameters can be set using
      // either the accessor functions of the L1InitManager class, or reading the beforehand prepared L1Parameters
      // object. On top of that, the parameters configuration (including cuts, algorithm execution scenario etc.) can
      // be initialized from the YAML configuration files in two levels (base and user). The base level of the initia-
      // lization is required for CbmL1 mode, the user level of the initialization is optional.

      std::string mainConfig = std::string(gSystem->Getenv("VMCWORKDIR")) + "/macro/L1/configs/";
      switch (fTrackingMode) {
        case L1Algo::TrackingMode::kSts: mainConfig += "ca_params_sts.yaml"; break;
        case L1Algo::TrackingMode::kMcbm: mainConfig += "ca_params_mcbm.yaml"; break;
        case L1Algo::TrackingMode::kGlobal: mainConfig += "ca_params_global.yaml"; break;
      }
      fInitManager.SetConfigMain(mainConfig);
    }

    fInitManager.SetDetectorNames(cbm::ca::kDetName);
    L1_SHOW(fInitManager.GetDetectorName(L1DetectorID::kSts));

    auto mvdInterface  = CbmMvdTrackingInterface::Instance();
    auto stsInterface  = CbmStsTrackingInterface::Instance();
    auto muchInterface = CbmMuchTrackingInterface::Instance();
    auto trdInterface  = CbmTrdTrackingInterface::Instance();
    auto tofInterface  = CbmTofTrackingInterface::Instance();

    fNMvdStationsGeom  = (fUseMVD) ? mvdInterface->GetNtrackingStations() : 0;
    fNStsStationsGeom  = (fUseSTS) ? stsInterface->GetNtrackingStations() : 0;
    fNMuchStationsGeom = (fUseMUCH) ? muchInterface->GetNtrackingStations() : 0;
    fNTrdStationsGeom  = (fUseTRD) ? trdInterface->GetNtrackingStations() : 0;
    fNTofStationsGeom  = (fUseTOF) ? tofInterface->GetNtrackingStations() : 0;
    fNStationsGeom = fNMvdStationsGeom + fNStsStationsGeom + fNMuchStationsGeom + fNTrdStationsGeom + fNTofStationsGeom;

    // **************************
    // ** Field initialization **
    // **************************

    if (FairRunAna::Instance()->GetField()) {
      fInitManager.SetFieldFunction([](const double(&inPos)[3], double(&outB)[3]) {
        FairRunAna::Instance()->GetField()->GetFieldValue(inPos, outB);
      });
    }
    else {
      fInitManager.SetFieldFunction([](const double(&)[3], double(&outB)[3]) {
        outB[0] = 0.;
        outB[1] = 0.;
        outB[2] = 0.;
      });
    }

    // ***************************
    // ** Target initialization **
    // ***************************

    GetTargetInfo();

    fInitManager.SetTargetPosition(fTargetX, fTargetY, fTargetZ);

    // *********************************
    // ** Target field initialization **
    // *********************************

    // FIXME: SZh 22.08.2023:
    //   Is there anyway to calculate a step between field slices?
    fInitManager.InitTargetField(/*zStep = */ 2.5 /*cm*/);  // Replace zStep -> sizeZfieldRegion = 2 * zStep (TODO)

    // **************************************
    // **                                  **
    // ** STATIONS GEOMETRY INITIALIZATION **
    // **                                  **
    // **************************************


    // ***************************************************
    // ** Active tracking detector subsystems selection **
    // ***************************************************

    std::set<L1DetectorID> vActiveTrackingDetectorIDs {};  // Set of detectors active in tracking

    if (fUseMVD) { vActiveTrackingDetectorIDs.insert(L1DetectorID::kMvd); }
    if (fUseSTS) { vActiveTrackingDetectorIDs.insert(L1DetectorID::kSts); }
    if (fUseMUCH) { vActiveTrackingDetectorIDs.insert(L1DetectorID::kMuch); }
    if (fUseTRD) { vActiveTrackingDetectorIDs.insert(L1DetectorID::kTrd); }
    if (fUseTOF) { vActiveTrackingDetectorIDs.insert(L1DetectorID::kTof); }
    fInitManager.SetActiveDetectorIDs(vActiveTrackingDetectorIDs);

    // *************************************
    // ** Stations layout initialization  **
    // *************************************

    std::vector<L1BaseStationInfo> vStations;
    vStations.reserve(100);

    // *** MVD stations info ***
    if (fUseMVD) {
      for (int iSt = 0; iSt < fNMvdStationsGeom; ++iSt) {
        auto stationInfo = L1BaseStationInfo(L1DetectorID::kMvd, iSt);
        // TODO: SZh 15.08.2022: Replace station type with L1DetectorID
        stationInfo.SetStationType(1);  // MVD
        stationInfo.SetTimeInfo(mvdInterface->IsTimeInfoProvided(iSt));
        stationInfo.SetFieldStatus(fTrackingMode == L1Algo::TrackingMode::kMcbm ? 0 : 1);
        stationInfo.SetZref(mvdInterface->GetZref(iSt));
        stationInfo.SetZmin(mvdInterface->GetZmin(iSt));
        stationInfo.SetZmax(mvdInterface->GetZmax(iSt));
        stationInfo.SetXmax(mvdInterface->GetXmax(iSt));
        stationInfo.SetYmax(mvdInterface->GetYmax(iSt));
        stationInfo.SetTrackingStatus(true);
        if (fvmDisabledStationIDs[L1DetectorID::kMvd].find(iSt) != fvmDisabledStationIDs[L1DetectorID::kMvd].cend()) {
          stationInfo.SetTrackingStatus(false);
        }
        fInitManager.AddStation(stationInfo);
        LOG(info) << "- MVD station " << iSt << " at z = " << stationInfo.GetZref() << " cm ";
      }
    }

    // *** STS stations info ***
    if (fUseSTS) {
      for (int iSt = 0; iSt < fNStsStationsGeom; ++iSt) {
        auto stationInfo = L1BaseStationInfo(L1DetectorID::kSts, iSt);
        // TODO: SZh 15.08.2022: Replace station type with L1DetectorID
        stationInfo.SetStationType(0);  // STS
        stationInfo.SetTimeInfo(stsInterface->IsTimeInfoProvided(iSt));
        stationInfo.SetTimeInfo(L1Algo::TrackingMode::kMcbm != fTrackingMode ? stsInterface->IsTimeInfoProvided(iSt)
                                                                             : false);
        stationInfo.SetFieldStatus(L1Algo::TrackingMode::kMcbm == fTrackingMode ? 0 : 1);
        stationInfo.SetZref(stsInterface->GetZref(iSt));
        stationInfo.SetZmin(stsInterface->GetZmin(iSt));
        stationInfo.SetZmax(stsInterface->GetZmax(iSt));
        stationInfo.SetXmax(stsInterface->GetXmax(iSt));
        stationInfo.SetYmax(stsInterface->GetYmax(iSt));

        stationInfo.SetTrackingStatus(true);
        if (fvmDisabledStationIDs[L1DetectorID::kSts].find(iSt) != fvmDisabledStationIDs[L1DetectorID::kSts].cend()) {
          stationInfo.SetTrackingStatus(false);
        }
        fInitManager.AddStation(stationInfo);
        LOG(info) << "- STS station " << iSt << " at z = " << stationInfo.GetZref() << " cm ";
      }
    }

    // *** MuCh stations info ***
    if (fUseMUCH) {
      for (int iSt = 0; iSt < fNMuchStationsGeom; ++iSt) {
        auto stationInfo = L1BaseStationInfo(L1DetectorID::kMuch, iSt);
        // TODO: SZh 15.08.2022: Replace station type with L1DetectorID
        stationInfo.SetStationType(2);  // MuCh
        stationInfo.SetTimeInfo(muchInterface->IsTimeInfoProvided(iSt));
        stationInfo.SetTimeInfo(L1Algo::TrackingMode::kMcbm != fTrackingMode ? stsInterface->IsTimeInfoProvided(iSt)
                                                                             : false);
        stationInfo.SetFieldStatus(0);
        stationInfo.SetZref(muchInterface->GetZref(iSt));
        stationInfo.SetZmin(muchInterface->GetZmin(iSt));
        stationInfo.SetZmax(muchInterface->GetZmax(iSt));
        stationInfo.SetXmax(muchInterface->GetXmax(iSt));
        stationInfo.SetYmax(muchInterface->GetYmax(iSt));

        stationInfo.SetTrackingStatus(true);
        if (fvmDisabledStationIDs[L1DetectorID::kMuch].find(iSt) != fvmDisabledStationIDs[L1DetectorID::kMuch].cend()) {
          stationInfo.SetTrackingStatus(false);
        }
        fInitManager.AddStation(stationInfo);
        LOG(info) << "- MuCh station " << iSt << " at z = " << stationInfo.GetZref() << " cm";
      }
    }

    // *** TRD stations info ***
    if (fUseTRD) {
      for (int iSt = 0; iSt < fNTrdStationsGeom; ++iSt) {
        auto stationInfo = L1BaseStationInfo(L1DetectorID::kTrd, iSt);
        // TODO: SZh 15.08.2022: Replace station type with L1DetectorID
        stationInfo.SetStationType((iSt == 1 || iSt == 3) ? 6 : 3);  // MuCh
        stationInfo.SetTimeInfo(trdInterface->IsTimeInfoProvided(iSt));
        stationInfo.SetTimeInfo(L1Algo::TrackingMode::kMcbm != fTrackingMode ? stsInterface->IsTimeInfoProvided(iSt)
                                                                             : false);
        stationInfo.SetFieldStatus(0);
        stationInfo.SetZref(trdInterface->GetZref(iSt));
        stationInfo.SetZmin(trdInterface->GetZmin(iSt));
        stationInfo.SetZmax(trdInterface->GetZmax(iSt));
        stationInfo.SetXmax(trdInterface->GetXmax(iSt));
        stationInfo.SetYmax(trdInterface->GetYmax(iSt));

        if (L1Algo::TrackingMode::kGlobal == fTrackingMode) { stationInfo.SetTimeInfo(false); }
        stationInfo.SetTrackingStatus(true);
        if (fvmDisabledStationIDs[L1DetectorID::kTrd].find(iSt) != fvmDisabledStationIDs[L1DetectorID::kTrd].cend()) {
          stationInfo.SetTrackingStatus(false);
        }
        fInitManager.AddStation(stationInfo);
        LOG(info) << "- TRD station " << iSt << " at z = " << stationInfo.GetZref() << " cm";
      }
    }

    // *** TOF stations info ***
    if (fUseTOF) {
      for (int iSt = 0; iSt < fNTofStationsGeom; ++iSt) {
        auto stationInfo = L1BaseStationInfo(L1DetectorID::kTof, iSt);
        // TODO: SZh 15.08.2022: Replace station type with L1DetectorID
        stationInfo.SetStationType(4);
        stationInfo.SetTimeInfo(tofInterface->IsTimeInfoProvided(iSt));
        stationInfo.SetTimeInfo(L1Algo::TrackingMode::kMcbm != fTrackingMode ? stsInterface->IsTimeInfoProvided(iSt)
                                                                             : false);
        stationInfo.SetFieldStatus(0);
        stationInfo.SetZref(tofInterface->GetZref(iSt));
        stationInfo.SetZmin(tofInterface->GetZmin(iSt));
        stationInfo.SetZmax(tofInterface->GetZmax(iSt));
        stationInfo.SetXmax(tofInterface->GetXmax(iSt));
        stationInfo.SetYmax(tofInterface->GetYmax(iSt));

        stationInfo.SetTrackingStatus(true);
        if (fvmDisabledStationIDs[L1DetectorID::kTof].find(iSt) != fvmDisabledStationIDs[L1DetectorID::kTof].cend()) {
          stationInfo.SetTrackingStatus(false);
        }
        fInitManager.AddStation(stationInfo);
        LOG(info) << "- TOF station " << iSt << " at z = " << stationInfo.GetZref() << " cm";
      }
    }

    // Init station layout: sort stations in z-position and init maps of station local, geo and active indices
    fInitManager.InitStationLayout();

    // ****************************************
    // ** Material maps initialization       **
    // ****************************************
    this->GenerateMaterialMaps();

    // *******************************
    // ** Initialize search windows **
    // *******************************

    if (fsInputSearchWindowsFilename.size()) {
      fInitManager.DevSetIsParSearchWUsed();
      fInitManager.ReadSearchWindows(fsInputSearchWindowsFilename);
    }
    else {
      fInitManager.DevSetIsParSearchWUsed(false);
    }

    // Form parameters container
    if (!fInitManager.FormParametersContainer()) { return kFATAL; }

    // Write parameters object to file if needed
    if (1 == fSTAPDataMode || 4 == fSTAPDataMode) { this->WriteSTAPParamObject(); }
  }


  // Init L1 algo core

  // FIXME: SZh 22.08.2023:
  //   Re-organize the the relation between CbmL1 and L1Algo. I belive, we don't need a global pointer to the L1Algo
  //   instance.
  fpAlgo = &gAlgo;
  fpAlgo->Init(fTrackingMode);

  //
  // ** Send formed parameters object to L1Algo instance **
  //
  fInitManager.SendParameters(fpAlgo);

  /*** Get numbers of active stations ***/

  fNMvdStations  = fpAlgo->GetParameters()->GetNstationsActive(L1DetectorID::kMvd);
  fNStsStations  = fpAlgo->GetParameters()->GetNstationsActive(L1DetectorID::kSts);
  fNTrdStations  = fpAlgo->GetParameters()->GetNstationsActive(L1DetectorID::kTrd);
  fNMuchStations = fpAlgo->GetParameters()->GetNstationsActive(L1DetectorID::kMuch);
  fNTofStations  = fpAlgo->GetParameters()->GetNstationsActive(L1DetectorID::kTof);
  fNStations     = fpAlgo->GetParameters()->GetNstationsActive();

  LOG(info) << fpAlgo->GetParameters()->ToString(1);
  fpIODataManager->SetNofActiveStations(fNStations);

  LOG(info) << "----- Numbers of stations active in tracking -----";
  LOG(info) << "  MVD:    " << fNMvdStations;
  LOG(info) << "  STS:    " << fNStsStations;
  LOG(info) << "  MuCh:   " << fNMuchStations;
  LOG(info) << "  TRD:    " << fNTrdStations;
  LOG(info) << "  ToF:    " << fNTofStations;
  LOG(info) << "  Total:  " << fNStations;

  // monitor the material
  {
    LOG(info) << "\033[31;1m-------------------- L1 material -----------------------------\033[0m";
    fMaterialMonitor.clear();
    for (int i = 0; i < fNStations; i++) {
      L1MaterialMonitor m(&(fpAlgo->GetParameters()->GetThicknessMaps()[i]), Form("station %d", i));
      LOG(info) << m.ToString();
      fMaterialMonitor.push_back(m);
    }
    LOG(info) << "\033[31;1m-------------------------------------------------------------\033[0m";
  }

  DumpMaterialToFile("L1material.root");

  // Initialize counters
  fEventNo       = 0;
  fNofRecoTracks = 0;

  return kSUCCESS;
}


void CbmL1::Reconstruct(CbmEvent* event)
{
  fvSelectedMcEvents.clear();

  int bestMcFile  = -1;
  int bestMcEvent = -1;

  if (fPerformance) {
    if (event) {
      CbmMatch* match = event->GetMatch();
      if (!match) {
        LOG(error) << "CbmL1: match between reco and MC events missing!! Performance can not be evaluated!!";
        fPerformance = 0;
      }
      else {
        cout << "CbmL1: mc events all " << fpMcEventList->GetNofEvents() << " mc events linked " << match->GetNofLinks()
             << std::endl;
        for (int iLink = 0; iLink < match->GetNofLinks(); iLink++) {
          const CbmLink& link = match->GetLink(iLink);
          fvSelectedMcEvents.insert(DFSET::value_type(link.GetFile(), link.GetEntry()));
          cout << "CbmL1: linked mc event file " << link.GetFile() << " event  " << link.GetEntry() << std::endl;
        }
        if (match->GetNofLinks() > 1) {
          const CbmLink& link = match->GetMatchedLink();
          bestMcFile          = link.GetFile();
          bestMcEvent         = link.GetEntry();
        }
      }
    }
    else {
      int nofEvents = fpMcEventList->GetNofEvents();
      for (int iE = 0; iE < nofEvents; iE++) {
        int fileId  = fpMcEventList->GetFileIdByIndex(iE);
        int eventId = fpMcEventList->GetEventIdByIndex(iE);
        fvSelectedMcEvents.insert(DFSET::value_type(fileId, eventId));
      }
    }
  }

  if (fVerbose > 1) { cout << "\nCbmL1::Exec event " << fEventNo << " ...\n\n"; }
#ifdef _OPENMP
  omp_set_num_threads(1);
#endif

  // ----- Read data from branches and send data from IODataManager to L1Algo ----------------------------------------

  ReadEvent(event);
  //if (fPerformance) {
  //  LOG(info) << "**** DEBUG: BEGIN";
  //  LOG(info) << fvMCPoints[0].ToString(10, true); // header
  //  for (const auto& point: fvMCPoints) { LOG(info) << point.ToString(10); }
  //  LOG(info) << "**** DEBUG: END";
  //}


  // Material monitoring: mark active areas
  {
    for (L1HitIndex_t i = 0; i < fpAlgo->GetInputData().GetNhits(); i++) {
      const L1Hit& h = fpAlgo->GetInputData().GetHit(i);
      fMaterialMonitor[h.iSt].MarkActiveBin(h.x, h.y);
    }
  }
  //

  if (fPerformance) {
    HitMatch();
    // calculate the max number of Hits\mcPoints on continuous(consecutive) stations
    for (auto it = fvMCTracks.begin(); it != fvMCTracks.end(); ++it) {
      it->Init();
    }
    if (bestMcFile >= 0) {  // suppress mc tracks from complementary mc events
      for (auto it = fvMCTracks.begin(); it != fvMCTracks.end(); ++it) {
        if (it->iFile != bestMcFile || it->iEvent != bestMcEvent) { it->SetIsReconstructable(false); }
      }
    }
  }

  //  FieldApproxCheck();
  //  FieldIntegralCheck();

  if (fVerbose > 1) { LOG(info) << "L1 Track finder..."; }
  fpAlgo->CaTrackFinder();
  //       IdealTrackFinder();
  fTrackingTime = fpAlgo->fCaRecoTime;

  if (fVerbose > 1) { LOG(info) << "L1 Track finder ok"; }

  // save reconstructed tracks

  fvRecoTracks.clear();
  fvRecoTracks.reserve(fpAlgo->fRecoTracks.size());

  int trackFirstHit = 0;

  for (L1Vector<L1Track>::iterator it = fpAlgo->fRecoTracks.begin(); it != fpAlgo->fRecoTracks.end();
       trackFirstHit += it->NHits, it++) {

    CbmL1Track t;

    for (int i = 0; i < L1TrackPar::kNparTr; i++) {
      t.T[i]     = it->TFirst[i];
      t.TLast[i] = it->TLast[i];
      t.Tpv[i]   = it->Tpv[i];
    }

    for (int i = 0; i < L1TrackPar::kNparCov; ++i) {
      t.C[i]     = it->CFirst[i];
      t.CLast[i] = it->CLast[i];
      t.Cpv[i]   = it->Cpv[i];
    }

    t.chi2 = it->chi2;
    t.NDF  = it->NDF;
    t.Hits.clear();
    t.mass        = fpAlgo->fDefaultMass;  // pion mass
    t.is_electron = 0;
    t.SetId(fvRecoTracks.size());

    for (int i = 0; i < it->NHits; i++) {
      int caHitId  = fpAlgo->fRecoHits[trackFirstHit + i];
      int cbmHitID = fpAlgo->GetInputData().GetHit(caHitId).ID;
      t.Hits.push_back(cbmHitID);
    }
    fvRecoTracks.push_back(t);
  }

  fNofRecoTracks += static_cast<int>(fpAlgo->fRecoTracks.size());

  LOG(debug) << "CA Track Finder: " << fpAlgo->fCaRecoTime << " s/sub-ts";


  // output performance
  if (fPerformance) {
    if (fVerbose > 1) { LOG(info) << "Performance..."; }

    //
    // tracker input performance is moved to external QA tasks.
    // InputPerformance() method is not optimised to run with the event builder
    // TODO: verify QA tasks and remove InputPerformance()
    // InputPerformance();
    //

    TrackMatch();
    EfficienciesPerformance();
    HistoPerformance();
    TrackFitPerformance();
    if (fsMcTripletsOutputFilename.size()) { DumpMCTripletsToTree(); }
    // TimeHist();
    ///    WriteSIMDKFData();
  }
  if (fVerbose > 1) { LOG(info) << "Tracking performance... done"; }
  if (fVerbose > 1) { LOG(info) << "End of CA"; }

  static bool ask = 0;
  char symbol;
  if (ask) {
    LOG(info) << "\nCA run (any/r/q) > ";
    do {
      std::cin.get(symbol);
      if (symbol == 'r') ask = false;
      if ((symbol == 'e') || (symbol == 'q')) exit(0);
    } while (symbol != '\n');
  }

  ++fEventNo;
}

// -----   Finish CbmStsFitPerformanceTask task   -----------------------------
void CbmL1::Finish()
{

  // monitor the material
  LOG(info) << "\033[31;1m ****************************\033[0m";
  LOG(info) << "\033[31;1m **  CA Tracking monitore  **\033[0m";
  LOG(info) << "\033[31;1m ****************************\033[0m";


  LOG(info) << "\033[31;1m ----- Material budget -------------------------------------- \033[0m";
  for (int i = 0; i < fNStations; i++) {
    LOG(info) << fMaterialMonitor[i].ToString();
  }
  LOG(info) << "\033[31;1m -------------------------------------------------------------\033[0m";

  // monitor of the reconstructed tracks
  LOG(info) << "\033[31;1m ----- Counters ----------------------------------------------\033[0m";
  LOG(info) << "\tNumber of analyzed events:                   " << fEventNo;
  LOG(info) << "\tNumber of reconstructed tracks:              " << fNofRecoTracks;
  LOG(info) << "\tNumber of reconstructed tracks per event/TS: " << static_cast<double>(fNofRecoTracks) / fEventNo;
  LOG(info) << "\033[31;1m -------------------------------------------------------------\033[0m";

  TDirectory* curr   = gDirectory;
  TFile* currentFile = gFile;


  // Open output file and write histograms
  boost::filesystem::path p = (FairRunAna::Instance()->GetUserOutputFileName()).Data();
  std::string dir           = p.parent_path().string();
  if (dir.empty()) dir = ".";
  {
    std::string histoOutName = dir + "/L1_histo_" + p.filename().string();
    LOG(info) << "\033[31;1mL1 performance histograms will be saved to: \033[0m" << histoOutName;
    TFile* outfile = new TFile(histoOutName.c_str(), "RECREATE");
    outfile->cd();
    writedir2current(fHistoDir);
    outfile->Close();
    outfile->Delete();
  }
  {
    std::string tablesOutName = dir + "/L1_perftable_" + p.filename().string();
    LOG(info) << "\033[31;1mL1 performance tables will be saved to: \033[0m" << tablesOutName;
    TFile* outfile = new TFile(tablesOutName.c_str(), "RECREATE");
    outfile->cd();
    writedir2current(fTableDir);
    outfile->Close();
    outfile->Delete();
  }

  if (fpMcTripletsOutFile) {
    fpMcTripletsOutFile->cd();
    fpMcTripletsTree->Write();
    fpMcTripletsOutFile->Close();
    fpMcTripletsOutFile->Delete();
  }

  gFile      = currentFile;
  gDirectory = curr;
  fpAlgo->Finish();
}


// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::writedir2current(TObject* obj)
{
  if (!obj->IsFolder()) obj->Write();
  else {
    TDirectory* cur = gDirectory;
    TDirectory* sub = cur->mkdir(obj->GetName());
    sub->cd();
    TList* listSub = (L1_DYNAMIC_CAST<TDirectory*>(obj))->GetList();
    TIter it(listSub);
    while (TObject* obj1 = it())
      writedir2current(obj1);
    cur->cd();
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::GenerateMaterialMaps()
{
  LOG(info) << "Generating material maps...";
  auto timerStart = std::chrono::high_resolution_clock::now();

  ca::tools::MaterialHelper matHelper;
  matHelper.SetSafeMaterialInitialization(fDoSafeMaterialInitialization);

  if (!fMatBudgetParallelProjection) { matHelper.SetDoRadialProjection(fTargetZ); }
  matHelper.SetNraysPerDim(fMatBudgetNrays);

  double zLast = fTargetZ + 1.;  // some gap (+1cm) to skip the target material

  auto& vStations = fInitManager.GetStationInfo();
  for (unsigned int ist = 0; ist < vStations.size(); ist++) {
    auto& station = vStations[ist];
    double z1     = station.GetZmax();
    double z2     = z1;
    if (ist < vStations.size() - 1) {
      // split materials between the stations at the middle
      auto& stationNext = vStations[ist + 1];
      z2                = stationNext.GetZmin();
    }
    double zNew = 0.5 * (z1 + z2);

    double maxXY = 1.3 * std::max(station.GetXmax(), station.GetYmax());
    //double maxXY = 80;
    // calculate n bins from the minimal pitch
    int nBins = static_cast<int>(std::ceil(2. * maxXY / fMatBudgetPitch));
    if (nBins < 1) { LOG(fatal) << " material nBins " << nBins << " is not positive, something is wrong"; }
    if (nBins > fMatBudgetNbins) { nBins = fMatBudgetNbins; }

    L1Material matBudget = matHelper.GenerateMaterialMap(station.GetZref(), zLast, zNew, maxXY, nBins);

    station.SetMaterialMap(matBudget);

    LOG(info) << "Generated material map for tracking station " << ist << " at z = " << station.GetZref() << " cm."
              << " Material is collected between z = " << zLast << " and z = " << zNew;

    zLast = zNew;
  }

  auto timerEnd                 = std::chrono::high_resolution_clock::now();
  double materialGenerationTime = (double) (std::chrono::duration<double>(timerEnd - timerStart).count());
  LOG(info) << "Generating material maps... done! (it took " << materialGenerationTime << " s)";
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::IdealTrackFinder()
{
  fpAlgo->fRecoTracks.clear();
  fpAlgo->fRecoHits.clear();

  for (L1Vector<CbmL1MCTrack>::iterator i = fvMCTracks.begin(); i != fvMCTracks.end(); ++i) {
    CbmL1MCTrack& MC = *i;

    if (!MC.IsReconstructable()) continue;
    if (!(MC.ID >= 0)) continue;

    if (MC.Hits.size() < 4) continue;

    L1Track algoTr;
    algoTr.NHits = 0;

    L1Vector<int> hitIndices("CbmL1::hitIndices", fpAlgo->GetParameters()->GetNstationsActive(), -1);

    for (unsigned int iH = 0; iH < MC.Hits.size(); iH++) {
      const int hitI               = MC.Hits[iH];
      const CbmL1HitDebugInfo& hit = fvHitDebugInfo[hitI];
      const int iP                 = hit.GetMCPointIndex();
      assert(iP >= 0);
      const int iStation = fvMCPoints[iP].iStation;

      if (iStation >= 0) hitIndices[iStation] = hitI;
    }


    for (int iH = 0; iH < fpAlgo->GetParameters()->GetNstationsActive(); iH++) {
      const int hitI = hitIndices[iH];
      if (hitI < 0) continue;

      // fpAlgo->fRecoHits.push_back(hitI);
      algoTr.NHits++;
    }


    if (algoTr.NHits < 3) continue;

    for (int iH = 0; iH < fpAlgo->GetParameters()->GetNstationsActive(); iH++) {
      const int hitI = hitIndices[iH];
      if (hitI < 0) continue;
      fpAlgo->fRecoHits.push_back(hitI);
    }

    algoTr.TFirst[0] = MC.x;
    algoTr.TFirst[1] = MC.y;
    algoTr.TFirst[2] = MC.px / MC.pz;
    algoTr.TFirst[3] = MC.py / MC.pz;
    algoTr.TFirst[4] = MC.q / MC.p;
    algoTr.TFirst[5] = MC.z;

    fpAlgo->fRecoTracks.push_back(algoTr);
  }
}  // void CbmL1::IdealTrackFinder()


/// -----   STandAlone Package service-functions  -----------------------------

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::DefineSTAPNames(const char* dirName)
{
  // FIXME: SZh 01.03.2023: Clean STAP names
  namespace bfs = boost::filesystem;

  if (fSTAPDataMode == 0) { return; }

  // Define file prefix (/path/to/data/setup.reco.root -> "setup.reco")
  bfs::path pathToRecoOutput = FairRunAna::Instance()->GetUserOutputFileName().Data();
  fSTAPDataPrefix            = pathToRecoOutput.filename().string();
  fSTAPDataPrefix.ReplaceAll(".root", "");
  fSTAPDataPrefix = fSTAPDataPrefix.Strip(TString::EStripType::kBoth, '.');

  TString sDirName = TString(dirName);
  if (sDirName.Length() == 0) {
    fSTAPDataDir = pathToRecoOutput.parent_path().string();
    L1_SHOW(fSTAPDataDir);
  }
  else if (bfs::exists(sDirName.Data()) && bfs::is_directory(sDirName.Data())) {
    fSTAPDataDir = sDirName;
  }
  else {
    fSTAPDataDir = ".";
  }

  LOG(info) << "CbmL1: STAP data root directory is \033[1;32m" << bfs::system_complete(fSTAPDataDir.Data())
            << "\033[0m";

  if (fSTAPDataMode == 4) { return; }

  // Directory for handling L1InputData objects
  TString sInputDataDir = fSTAPDataDir + "/" + fSTAPDataPrefix + "_cainputdata";
  if (!bfs::exists(sInputDataDir.Data())) {
    LOG(warn) << "CbmL1: directory for tracking input data does not exist. It will be created";
    bfs::create_directories(sInputDataDir.Data());
  }
  LOG(info) << "CbmL1: STAP tracking input jobs directory is \033[1;32m" << bfs::system_complete(sInputDataDir.Data())
            << "\033[0m";
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::WriteSTAPParamObject()
{
  TString filename = fSTAPDataDir + "/" + fSTAPDataPrefix + "." + TString(kSTAPParamSuffix.data());
  fInitManager.WriteParametersObject(filename.Data());
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::WriteSTAPAlgoInputData(int iJob)  // must be called after ReadEvent
{
  TString filename =
    fSTAPDataDir + "/" + fSTAPDataPrefix + "_cainputdata/" + +TString::Format(kSTAPAlgoIDataSuffix.data(), iJob);

  // Write file
  fpIODataManager->WriteInputData(filename.Data());
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::WriteSTAPPerfInputData()  // must be called after ReadEvent
{
  LOG(fatal) << "CbmL1: Running in standalone mode is not available at the moment. It will be updated soon...";
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::ReadSTAPParamObject()
{
  TString filename = fSTAPDataDir + "/" + fSTAPDataPrefix + "." + TString(kSTAPParamSuffix.data());
  fInitManager.ReadParametersObject(filename.Data());
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::ReadSTAPAlgoInputData(int iJob)
{
  TString filename = fSTAPDataDir + "/" + TString(kSTAPAlgoIDataDir) + "/" + fSTAPDataPrefix + "."
                     + TString::Format(kSTAPAlgoIDataSuffix.data(), iJob);

  // Read file
  fpIODataManager->ReadInputData(filename.Data());
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::ReadSTAPPerfInputData()
{
  LOG(fatal) << "CbmL1: Running in standalone mode is not available at the moment. It will be updated soon...";
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmL1::WriteSIMDKFData()
{
  // TODO: Must be totally reimplemented (S.Zharko)
  static bool first = 1;

  ///Write Tracks and MC Tracks

  static int TrNumber = 0;
  std::fstream Tracks, McTracksCentr, McTracksIn, McTracksOut;
  if (first) {
    Tracks.open("tracks.dat", std::fstream::out);
    McTracksCentr.open("mctrackscentr.dat", std::fstream::out);
    McTracksIn.open("mctracksin.dat", std::fstream::out);
    McTracksOut.open("mctracksout.dat", std::fstream::out);
    first = 0;
  }
  else {
    Tracks.open("tracks.dat", std::fstream::out | std::fstream::app);
    McTracksCentr.open("mctrackscentr.dat", std::fstream::out | std::fstream::app);
    McTracksIn.open("mctracksin.dat", std::fstream::out | std::fstream::app);
    McTracksOut.open("mctracksout.dat", std::fstream::out | std::fstream::app);
  }

  for (L1Vector<CbmL1Track>::iterator RecTrack = fvRecoTracks.begin(); RecTrack != fvRecoTracks.end(); ++RecTrack) {
    if (RecTrack->IsGhost()) continue;

    CbmL1MCTrack* MCTrack = RecTrack->GetMCTrack();
    if (!(MCTrack->IsPrimary())) continue;

    int NHits = (RecTrack->Hits).size();
    float x[20], y[20], z[20];
    int st[20];
    int jHit = 0;
    for (int iHit = 0; iHit < NHits; iHit++) {
      CbmL1HitDebugInfo& h = fvHitDebugInfo[RecTrack->Hits[iHit]];
      st[jHit]             = h.iStation;
      if (h.ExtIndex < 0) {
        CbmMvdHit* MvdH = (CbmMvdHit*) fpMvdHits->At(-h.ExtIndex - 1);
        x[jHit]         = MvdH->GetX();
        y[jHit]         = MvdH->GetY();
        z[jHit]         = MvdH->GetZ();
        jHit++;
      }
      else {
        CbmStsHit* StsH = (CbmStsHit*) fpStsHits->At(h.ExtIndex);
        x[jHit]         = StsH->GetX();
        y[jHit]         = StsH->GetY();
        z[jHit]         = StsH->GetZ();
        jHit++;
      }
    }

    Tracks << TrNumber << " " << MCTrack->x << " " << MCTrack->y << " " << MCTrack->z << " " << MCTrack->px << " "
           << MCTrack->py << " " << MCTrack->pz << " " << MCTrack->q << " " << NHits << endl;

    float AngleXAxis = 0, AngleYAxis = 0;
    for (int i = 0; i < NHits; i++)
      Tracks << "     " << st[i] << " " << x[i] << " " << y[i] << " " << z[i] << " " << AngleXAxis << " " << AngleYAxis
             << endl;
    Tracks << endl;

    int NMCPoints = (MCTrack->Points).size();

    McTracksIn << TrNumber << " " << MCTrack->x << " " << MCTrack->y << " " << MCTrack->z << " " << MCTrack->px << " "
               << MCTrack->py << " " << MCTrack->pz << " " << MCTrack->q << " " << NMCPoints << endl;
    McTracksOut << TrNumber << " " << MCTrack->x << " " << MCTrack->y << " " << MCTrack->z << " " << MCTrack->px << " "
                << MCTrack->py << " " << MCTrack->pz << " " << MCTrack->q << " " << NMCPoints << endl;
    McTracksCentr << TrNumber << " " << MCTrack->x << " " << MCTrack->y << " " << MCTrack->z << " " << MCTrack->px
                  << " " << MCTrack->py << " " << MCTrack->pz << " " << MCTrack->q << " " << NMCPoints << endl;

    for (int iPoint = 0; iPoint < NMCPoints; iPoint++) {
      CbmL1MCPoint& MCPoint = fvMCPoints[MCTrack->Points[iPoint]];
      McTracksIn << "     " << MCPoint.iStation << " " << MCPoint.xIn << " " << MCPoint.yIn << " " << MCPoint.zIn << " "
                 << MCPoint.pxIn << " " << MCPoint.pyIn << " " << MCPoint.pzIn << endl;
      McTracksOut << "     " << MCPoint.iStation << " " << MCPoint.xOut << " " << MCPoint.yOut << " " << MCPoint.zOut
                  << " " << MCPoint.pxOut << " " << MCPoint.pyOut << " " << MCPoint.pzOut << endl;
      McTracksCentr << "     " << MCPoint.iStation << " " << MCPoint.x << " " << MCPoint.y << " " << MCPoint.z << " "
                    << MCPoint.px << " " << MCPoint.py << " " << MCPoint.pz << endl;
    }
    McTracksIn << endl;
    McTracksOut << endl;
    McTracksCentr << endl;

    TrNumber++;
  }
}


// ---------------------------------------------------------------------------------------------------------------------
//
double CbmL1::boundedGaus(double sigma)
{
  assert(sigma > 0. && std::isfinite(sigma));
  double x = 0.;
  do {
    x = gRandom->Gaus(0., sigma);
  } while (fabs(x) > 3.5 * sigma);
  return x;
}

void CbmL1::DumpMaterialToFile(TString fileName)
{
  TFile* f = new TFile(fileName, "RECREATE");
  f->cd();
  const L1Parameters& param        = *fpAlgo->GetParameters();
  const L1MaterialContainer_t& map = param.GetThicknessMaps();
  for (int ist = 0; ist < param.GetNstationsActive(); ist++) {
    //const L1Station& st   = param.GetStation(ist);
    const L1Material& mat = map[ist];

    TString name  = Form("tracking_station%d", ist);
    TString title = Form("Tracking station %d: Rad. thickness in [%s]. Z region [%.2f,%.2f] cm.", ist, "%",
                         mat.GetZmin(), mat.GetZmax());

    if (fMatBudgetParallelProjection) { title += " Horisontal projection."; }
    else {
      title += " Radial projection.";
    }

    TH2F* h = new TH2F(name.Data(), title.Data(), mat.GetNbins(), -mat.GetXYmax(), mat.GetXYmax(), mat.GetNbins(),
                       -mat.GetXYmax(), mat.GetXYmax());
    h->GetXaxis()->SetTitle("X [cm]");
    h->GetYaxis()->SetTitle("Y [cm]");

    for (int ix = 0; ix < mat.GetNbins(); ix++) {
      for (int iy = 0; iy < mat.GetNbins(); iy++) {
        h->SetBinContent(ix + 1, iy + 1, 100. * mat.GetRadThickBin(ix, iy));
      }
    }
  }
  f->Write();
}

// ---------------------------------------------------------------------------------------------------------------------
// Target finder
//

void CbmL1::GetTargetInfo()
{
  // Loop over all nodes till a node with name "target" is found
  // Extract the required infrmation from the node and store it in the
  // proper structure
  // The complete logic depends on the naming convention of the target.
  // If the node doesn't contain the string target the procedure will fail

  fTargetX = 1.e10;
  fTargetY = 1.e10;
  fTargetZ = 1.e10;

  TString targetPath;
  TGeoNode* targetNode {nullptr};
  FindTargetNode(targetPath, targetNode);

  if (!targetNode) { LOG(fatal) << "L1: can not find the target!"; }

  Double_t local[3] = {0., 0., 0.};  // target centre, local c.s.
  Double_t global[3];                // target centre, global c.s.
  gGeoManager->cd(targetPath);
  gGeoManager->GetCurrentMatrix()->LocalToMaster(local, global);
  fTargetX = global[0];
  fTargetY = global[1];
  fTargetZ = global[2];

  LOG(info) << " Target found: \"" << targetPath << "\" at ( " << fTargetX << " " << fTargetY << " " << fTargetZ
            << " ) ";
}

// ---------------------------------------------------------------------------------------------------------------------
// Target finder - recursive routine
//

void CbmL1::FindTargetNode(TString& targetPath, TGeoNode*& targetNode)
{
  if (!targetNode) {  // init at the top of the tree
    targetNode = gGeoManager->GetTopNode();
    targetPath = "/" + TString(targetNode->GetName());
  }

  if (TString(targetNode->GetName()).Contains("target")) { return; }

  for (Int_t iNode = 0; iNode < targetNode->GetNdaughters(); iNode++) {
    TGeoNode* newNode = targetNode->GetDaughter(iNode);
    TString newPath   = targetPath + "/" + newNode->GetName();
    FindTargetNode(newPath, newNode);
    if (newNode) {
      targetPath = newPath;
      targetNode = newNode;
      return;
    }
  }
  targetPath = "";
  targetNode = nullptr;
}
