/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

#include "L1Station.h"

#include <Logger.h>

#include <iomanip>
#include <sstream>

//------------------------------------------------------------------------------------------------------------------------------------
//
void L1Station::CheckConsistency() const
{
  /*
   * Integer fields initialization checks
   */

  if (type < 0) {
    std::stringstream msg;
    msg << "L1Station: station type was not initialized (type = " << type << ", type > 0)";
    throw std::logic_error(msg.str());
  }

  if (timeInfo != 0 && timeInfo != 1) {
    std::stringstream msg;
    msg << "L1Station: illegal time information flag (timeInfo = " << timeInfo << ", "
        << "0 [time information is not used] or 1 [time information is used] expected)";
    throw std::logic_error(msg.str());
  }

  if (fieldStatus != 0 && fieldStatus != 1) {
    std::stringstream msg;
    msg << "L1Station: illegal field status flag (fieldStatus = " << fieldStatus << ", "
        << "0 [station is outside the field] or 1 [station is inside the field] expected";
    throw std::logic_error(msg.str());
  }

  /*
   * SIMD vector checks: all the words in a SIMD vector must be equal
   */

  L1Utils::CheckSimdVectorEquality(fZ, "L1Station::fZ");
  L1Utils::CheckSimdVectorEquality(Xmax, "L1Station::Xmax");
  L1Utils::CheckSimdVectorEquality(Ymax, "L1Station::Ymax");

  /*
   * Inner and outer radia checks:
   *  (i)  both Xmax and Ymax must be > 0
   */

  if (Xmax[0] < 0. || Ymax[0] < 0.) {
    std::stringstream msg;
    msg << "L1Station: " << this->ToString() << " has incorrect sizes: "
        << "Xmax = " << Xmax[0] << " [cm], Ymax = " << Ymax[0] << " [cm] (0 < Xmax && 0 < Ymax expected)";
    throw std::logic_error(msg.str());
  }


  /*
   * Check consistency of other members
   */

  //materialInfo.CheckConsistency();
  // TODO: Temporarily switched off, because Much has RL = 0, which is actually incorrect, but really is not used.
  //       One should provide average radiation length values for each Much layer (S.Zharko)
  fieldSlice.CheckConsistency();
}

//------------------------------------------------------------------------------------------------------------------------------------
//
std::string L1Station::ToString(int verbosityLevel, int indentLevel, bool isHeader) const
{
  std::stringstream msg {};
  constexpr char indentChar = '\t';
  constexpr char columnSize = 15;
  std::string indent(indentLevel, indentChar);
  if (verbosityLevel == 0) { msg << "station type = " << type << ", z = " << fZ[0] << " cm"; }
  else {
    if (isHeader) {
      verbosityLevel = 0;
      msg << indent;
      msg << std::setw(columnSize) << std::setfill(' ') << "type" << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << "time status" << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << "field status" << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << "z [cm]" << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << "Ymax [cm]" << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << "Xmax [cm]";
    }
    else {
      msg << indent;
      msg << std::setw(columnSize) << std::setfill(' ') << this->GetType() << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << this->GetTimeStatus() << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << this->GetFieldStatus() << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << this->GetZScal() << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << this->GetXmaxScal() << ' ';
      msg << std::setw(columnSize) << std::setfill(' ') << this->GetYmaxScal();
    }
    if (verbosityLevel > 3) {
      msg << indent << "Field approcimation coefficients:\n";
      msg << fieldSlice.ToString(indentLevel + 1) << '\n';
    }
  }
  return msg.str();
}
