/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmCaInputQaSts.cxx
/// \date   13.01.2023
/// \brief  QA-task for CA tracking input from MuCh detector (implementation)
/// \author S.Zharko <s.zharko@gsi.de>

#include "CbmCaInputQaSts.h"

#include "CbmAddress.h"
#include "CbmMCDataArray.h"
#include "CbmMCEventList.h"
#include "CbmMCTrack.h"
#include "CbmMatch.h"
#include "CbmQaCanvas.h"
#include "CbmQaTable.h"
#include "CbmQaUtil.h"
#include "CbmStsCluster.h"
#include "CbmStsHit.h"
#include "CbmStsPoint.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTimeSlice.h"

#include "FairRootManager.h"
#include "Logger.h"

#include "TBox.h"
#include "TClonesArray.h"
#include "TEllipse.h"
#include "TF1.h"
#include "TFormula.h"
#include "TGraphAsymmErrors.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TMath.h"
#include "TProfile.h"
#include "TProfile2D.h"
#include "TStyle.h"

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <numeric>
#include <tuple>

#include "L1Constants.h"

ClassImp(CbmCaInputQaSts);

namespace phys = L1Constants::phys;  // from L1Constants.h

// ---------------------------------------------------------------------------------------------------------------------
//
CbmCaInputQaSts::CbmCaInputQaSts(int verbose, bool isMCUsed) : CbmQaTask("CbmCaInputQaSts", verbose, isMCUsed)
{
  // TODO: remove an enlarged threshold when the pulls are fixed
  SetPullMeanDeviationThrsh(0.1);
  SetPullWidthDeviationThrsh(2.);
  SetEfficiencyThrsh(0.5, 0, 100);
}

// ---------------------------------------------------------------------------------------------------------------------
//
bool CbmCaInputQaSts::Check()
{
  bool res = true;

  int nSt = fpDetInterface->GetNtrackingStations();


  // **************************************************************
  // ** Basic checks, available both for real and simulated data **
  // **************************************************************

  // Fill efficiency distributions

  for (int iSt = 0; iSt < nSt; ++iSt) {
    TProfile2D* effxy = fvpe_reco_eff_vs_xy[iSt];
    for (int i = 1; i < effxy->GetNbinsX() - 1; i++) {
      for (int j = 1; j < effxy->GetNbinsY() - 1; j++) {
        int bin = effxy->GetBin(i, j);
        if (effxy->GetBinEntries(bin) >= 1) {
          fvph_reco_eff[iSt]->Fill(effxy->GetBinContent(bin));
          fvph_reco_eff[nSt]->Fill(effxy->GetBinContent(bin));
        }
      }
    }
  }

  // ----- Checks for mismatches in the ordering of the stations
  //
  std::vector<double> vStationPos(nSt, 0.);
  for (int iSt = 0; iSt < nSt; ++iSt) {
    vStationPos[iSt] = fpDetInterface->GetZ(iSt);
  }

  if (!std::is_sorted(vStationPos.cbegin(), vStationPos.cend(), [](int l, int r) { return l <= r; })) {
    if (fVerbose > 0) {
      LOG(error) << fName << ": stations are ordered improperly along the beam axis:";
      for (auto z : vStationPos) {
        LOG(error) << "\t- " << z;
      }
    }
    res = false;
  }

  // ----- Checks for mismatch between station and hit z positions
  //   The purpose of this block is to be ensured, that hits belong to the correct tracking station. For each tracking
  // station a unified position z-coordinate is defined, which generally differs from the corresponding positions of
  // reconstructed hits. This is due to non-regular position along the beam axis for each detector sensor. Thus, the
  // positions of the hits along z-axis are distributed relatively to the position of the station in some range.
  // If hits goes out this range, it can signal about a mismatch between hits and geometry. For limits of the range
  // one should select large enough values to cover the difference described above and in the same time small enough
  // to avoid overlaps with the neighboring stations.
  //   For each station, a distribution of z_{hit} - z_{st} is integrated over the defined range and scaled by the
  // total number of entries to the distribution. If this value is smaller then unity, then some of the hits belong to
  // another station.
  //
  for (int iSt = 0; iSt < nSt; ++iSt) {
    int nHits = fvph_hit_station_delta_z[iSt]->GetEntries();
    if (!nHits) {
      LOG_IF(error, fVerbose > 0) << fName << ": station " << iSt << " does not have hits";
      res = false;
      continue;
    }
    int iBinMin = fvph_hit_station_delta_z[iSt]->FindBin(-fMaxDiffZStHit);
    int iBinMax = fvph_hit_station_delta_z[iSt]->FindBin(+fMaxDiffZStHit);

    if (fvph_hit_station_delta_z[iSt]->Integral(iBinMin, iBinMax) < nHits) {
      LOG_IF(error, fVerbose > 0) << fName << ": station " << iSt << " has mismatches in hit z-positions";
      res = false;
    }
  }


  // *******************************************************
  // ** Additional checks, if MC information is available **
  // *******************************************************

  if (IsMCUsed()) {
    // ----- Check efficiencies
    // Error is raised, if any station has integrated efficiency lower then a selected threshold.
    // Integrated efficiency is estimated with fitting the efficiency curve for hit distances (r) with a uniform
    // distribution in the range from kEffRangeMin to kEffRangeMax, where the efficiency is assigned to be constant
    //
    // Fit efficiency curves
    LOG(info) << "-- Hit efficiency integrated over hit distance from station center";

    auto* pEffTable = MakeQaObject<CbmQaTable>("vs Station/eff_table", "Efficiency table", nSt, 2);
    pEffTable->SetNamesOfCols({"Station ID", "Efficiency"});
    pEffTable->SetColWidth(20);

    for (int iSt = 0; iSt < nSt; ++iSt) {
      auto eff = fvph_reco_eff[iSt]->GetMean();
      pEffTable->SetCell(iSt, 0, iSt);
      pEffTable->SetCell(iSt, 1, eff);
      res = CheckRange("Hit finder efficiency in station " + std::to_string(iSt), eff, fEffThrsh, 1.000);
    }

    LOG(info) << '\n' << pEffTable->ToString(3);

    // ----- Checks for residuals
    // Check hits for potential biases from central values

    auto* pResidualsTable =
      MakeQaObject<CbmQaTable>("vs Station/residuals_mean", "Residual mean values in different stations", nSt, 4);
    pResidualsTable->SetNamesOfCols({"Station ID", "Residual(x) [cm]", "Residual(y) [cm]", "Residual(t) [ns]"});
    pResidualsTable->SetColWidth(20);

    // Fit residuals
    for (int iSt = 0; iSt <= nSt; ++iSt) {

      cbm::qa::util::SetLargeStats(fvph_res_x[iSt]);
      cbm::qa::util::SetLargeStats(fvph_res_y[iSt]);
      cbm::qa::util::SetLargeStats(fvph_res_t[iSt]);
      cbm::qa::util::SetLargeStats(fvph_res_u[iSt]);
      cbm::qa::util::SetLargeStats(fvph_res_v[iSt]);

      pResidualsTable->SetCell(iSt, 0, iSt);
      pResidualsTable->SetCell(iSt, 1, fvph_res_x[iSt]->GetStdDev());
      pResidualsTable->SetCell(iSt, 2, fvph_res_y[iSt]->GetStdDev());
      pResidualsTable->SetCell(iSt, 3, fvph_res_t[iSt]->GetStdDev());
    }

    LOG(info) << '\n' << pResidualsTable->ToString(8);

    // ----- Checks for pulls
    //
    // For the hit pull is defined as a ration of the difference between hit and MC-point position or time component
    // values and an error of the hit position (or time) component, calculated in the same z-positions. In ideal case,
    // when the resolutions are well determined for detecting stations, the pull distributions should have a RMS equal
    // to unity. Here we allow a RMS of the pull distributions to be varied in a predefined range. If the RMS runs out
    // this range, QA task fails.

    auto* pPullsTable =
      MakeQaObject<CbmQaTable>("vs Station/pulls_rms", "Pulls std. dev. values in different stations", nSt, 4);
    pPullsTable->SetNamesOfCols({"Station ID", "Pull(x) sigm", "Pull(y) sigm", "Pull(t) sigm"});
    pPullsTable->SetColWidth(20);

    for (int iSt = 0; iSt < nSt + 1; ++iSt) {

      // Fit pull distributions for nicer representation. Fit results are not used in further checks.

      cbm::qa::util::SetLargeStats(fvph_pull_x[iSt]);
      cbm::qa::util::SetLargeStats(fvph_pull_y[iSt]);
      cbm::qa::util::SetLargeStats(fvph_pull_t[iSt]);
      cbm::qa::util::SetLargeStats(fvph_pull_u[iSt]);
      cbm::qa::util::SetLargeStats(fvph_pull_v[iSt]);

      // Check the pull quality
      res = CheckRangePull(fvph_pull_x[iSt]) && res;
      res = CheckRangePull(fvph_pull_y[iSt]) && res;
      res = CheckRangePull(fvph_pull_t[iSt]) && res;
      res = CheckRangePull(fvph_pull_u[iSt]) && res;
      res = CheckRangePull(fvph_pull_v[iSt]) && res;

      pPullsTable->SetCell(iSt, 0, iSt);
      pPullsTable->SetCell(iSt, 1, fvph_pull_x[iSt]->GetStdDev());
      pPullsTable->SetCell(iSt, 2, fvph_pull_y[iSt]->GetStdDev());
      pPullsTable->SetCell(iSt, 3, fvph_pull_t[iSt]->GetStdDev());
    }

    for (int idig = 0; idig <= fkMaxDigisInClusterForPulls; idig++) {
      cbm::qa::util::SetLargeStats(fvph_pull_u_Ndig[idig]);
      res = CheckRangePull(fvph_pull_u_Ndig[idig]) && res;

      cbm::qa::util::SetLargeStats(fvph_pull_v_Ndig[idig]);
      res = CheckRangePull(fvph_pull_v_Ndig[idig]) && res;
    }

    LOG(info) << '\n' << pPullsTable->ToString(3);
  }  // McUsed

  return res;
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaSts::DeInit()
{
  // Vectors with pointers to histograms
  fvph_hit_xy.clear();
  fvph_hit_zx.clear();
  fvph_hit_zy.clear();

  fvph_hit_station_delta_z.clear();

  fvph_hit_dx.clear();
  fvph_hit_dy.clear();
  fvph_hit_du.clear();
  fvph_hit_dv.clear();
  fvph_hit_kuv.clear();
  fvph_hit_dt.clear();

  fvph_n_points_per_hit.clear();

  fvph_point_xy.clear();
  fvph_point_zx.clear();
  fvph_point_zy.clear();

  fvph_point_hit_delta_z.clear();

  fvph_res_x.clear();
  fvph_res_y.clear();
  fvph_res_u.clear();
  fvph_res_v.clear();
  fvph_res_t.clear();

  fvph_pull_x.clear();
  fvph_pull_y.clear();
  fvph_pull_u.clear();
  fvph_pull_v.clear();
  fvph_pull_t.clear();

  fvph_pull_u_Ndig.clear();
  fvph_pull_v_Ndig.clear();

  fvph_res_x_vs_x.clear();
  fvph_res_y_vs_y.clear();
  fvph_res_u_vs_u.clear();
  fvph_res_v_vs_v.clear();
  fvph_res_t_vs_t.clear();

  fvph_pull_x_vs_x.clear();
  fvph_pull_y_vs_y.clear();
  fvph_pull_u_vs_u.clear();
  fvph_pull_v_vs_v.clear();
  fvph_pull_t_vs_t.clear();

  fvpe_reco_eff_vs_xy.clear();
  fvph_reco_eff.clear();
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaSts::FillHistograms()
{
  int nSt       = fpDetInterface->GetNtrackingStations();
  int nHits     = fpHits->GetEntriesFast();
  int nMCevents = (IsMCUsed()) ? fpMCEventList->GetNofEvents() : -1;

  std::vector<std::vector<std::vector<int>>>
    vNofHitsPerMcTrack;  // Number of hits per MC track per station in different MC events

  if (IsMCUsed()) {
    vNofHitsPerMcTrack.resize(nMCevents);
    for (int iE = 0; iE < nMCevents; ++iE) {
      int iFile     = fpMCEventList->GetFileIdByIndex(iE);
      int iEvent    = fpMCEventList->GetEventIdByIndex(iE);
      int nMcTracks = fpMCTracks->Size(iFile, iEvent);
      vNofHitsPerMcTrack[iE].resize(nSt);
      for (int iSt = 0; iSt < nSt; iSt++) {
        vNofHitsPerMcTrack[iE][iSt].resize(nMcTracks, 0);
      }
    }
  }

  for (int iH = 0; iH < nHits; ++iH) {
    const auto* pHit = dynamic_cast<const CbmStsHit*>(fpHits->At(iH));
    LOG_IF(fatal, !pHit) << fName << ": hit with iH = " << iH << " is not an CbmStsHit (dynamic cast failed)";

    // *************************
    // ** Reconstructed hit QA

    // Hit station geometry info
    int iSt = fpDetInterface->GetTrackingStationIndex(pHit);
    LOG_IF(fatal, iSt < 0 || iSt >= nSt) << fName << ": index of station (" << iSt << ") is out of range "
                                         << "for hit with id = " << iH;
    double stPhiU;  // STS front strips stereo angle
    double stPhiV;  // STS back strips stereo angle

    std::tie(stPhiU, stPhiV) = fpDetInterface->GetStereoAnglesSensor(pHit->GetAddress());

    double sinU = sin(stPhiU);
    double cosU = cos(stPhiU);
    double sinV = sin(stPhiV);
    double cosV = cos(stPhiV);

    // hit cov matrix
    double Cxx = pHit->GetDx() * pHit->GetDx();
    double Cxy = pHit->GetDxy();
    double Cyy = pHit->GetDy() * pHit->GetDy();

    // Hit position
    double xHit = pHit->GetX();
    double yHit = pHit->GetY();
    double zHit = pHit->GetZ();
    double uHit = xHit * cosU + yHit * sinU;
    double vHit = xHit * cosV + yHit * sinV;

    // Hit errors
    //double duHit = pHit->GetDu();
    //double dvHit = pHit->GetDv();
    double duHit = sqrt(cosU * cosU * Cxx + 2. * sinU * cosU * Cxy + sinU * sinU * Cyy);
    double dvHit = sqrt(cosV * cosV * Cxx + 2. * sinV * cosV * Cxy + sinV * sinV * Cyy);
    double dxHit = pHit->GetDx();
    double dyHit = pHit->GetDy();

    double duvHit = Cxx * cosU * cosV + Cxy * (sinU * cosV + cosU * sinV) + Cyy * sinU * sinV;
    double kuvHit = duvHit / duHit / dvHit;

    // Hit time
    double tHit  = pHit->GetTime();
    double dtHit = pHit->GetTimeError();

    fvph_hit_xy[iSt]->Fill(xHit, yHit);
    fvph_hit_zx[iSt]->Fill(zHit, xHit);
    fvph_hit_zy[iSt]->Fill(zHit, yHit);

    fvph_hit_station_delta_z[iSt]->Fill(zHit - fpDetInterface->GetZref(iSt));

    fvph_hit_dx[iSt]->Fill(dxHit);
    fvph_hit_dy[iSt]->Fill(dyHit);
    fvph_hit_du[iSt]->Fill(duHit);
    fvph_hit_dv[iSt]->Fill(dvHit);
    fvph_hit_kuv[iSt]->Fill(kuvHit);
    fvph_hit_dt[iSt]->Fill(dtHit);

    fvph_hit_xy[nSt]->Fill(xHit, yHit);
    fvph_hit_zx[nSt]->Fill(zHit, xHit);
    fvph_hit_zy[nSt]->Fill(zHit, yHit);

    fvph_hit_station_delta_z[nSt]->Fill(zHit - fpDetInterface->GetZref(iSt));

    fvph_hit_dx[nSt]->Fill(dxHit);
    fvph_hit_dy[nSt]->Fill(dyHit);
    fvph_hit_du[nSt]->Fill(duHit);
    fvph_hit_dv[nSt]->Fill(dvHit);
    fvph_hit_kuv[nSt]->Fill(kuvHit);
    fvph_hit_dt[nSt]->Fill(dtHit);

    // **********************
    // ** MC information QA

    if (IsMCUsed()) {
      const auto* pHitMatch = dynamic_cast<CbmMatch*>(fpHitMatches->At(iH));
      assert(pHitMatch);

      // Evaluate number of hits per one MC point
      int nMCpoints = 0;  // Number of MC points for this hit
      for (int iLink = 0; iLink < pHitMatch->GetNofLinks(); ++iLink) {
        const auto& link = pHitMatch->GetLink(iLink);

        int iP = link.GetIndex();  // Index of MC point

        // Skip noisy links
        if (iP < 0) { continue; }

        ++nMCpoints;

        int iE = fpMCEventList->GetEventIndex(link);

        LOG_IF(fatal, iE < 0 || iE >= nMCevents) << fName << ": id of MC event is out of range (hit id = " << iH
                                                 << ", link id = " << iLink << ", event id = " << iE << ')';

        // matched point
        const auto* pMCPoint = dynamic_cast<const CbmStsPoint*>(fpMCPoints->Get(link));
        LOG_IF(fatal, !pMCPoint) << fName << ": MC point object does not exist for hit " << iH;

        int iTr = pMCPoint->GetTrackID();

        LOG_IF(fatal, iTr >= (int) vNofHitsPerMcTrack[iE][iSt].size())
          << fName << ": index of MC track is out of range (hit id = " << iH << ", link id = " << iLink
          << ", event id = " << iE << ", track id = " << iTr << ')';

        vNofHitsPerMcTrack[iE][iSt][iTr]++;
      }

      fvph_n_points_per_hit[iSt]->Fill(nMCpoints);
      fvph_n_points_per_hit[nSt]->Fill(nMCpoints);

      //
      // Fill the following histograms exclusively for isolated hits with a one-to-one correspondence to the mc track
      // The mc track must satisfy the cuts
      //

      if (pHitMatch->GetNofLinks() != 1) { continue; }

      // The best link to in the match (probably, the cut on nMCpoints is meaningless)
      assert(pHitMatch->GetNofLinks() > 0);  // Should be always true due to the cut above
      const auto& bestPointLink = pHitMatch->GetMatchedLink();

      // Skip noisy links
      if (bestPointLink.GetIndex() < 0) { continue; }

      // Point matched by the best link
      const auto* pMCPoint = dynamic_cast<const CbmStsPoint*>(fpMCPoints->Get(bestPointLink));
      LOG_IF(fatal, !pMCPoint) << fName << ": MC point object does not exist for hit " << iH;

      // MC track for this point
      CbmLink bestTrackLink = bestPointLink;
      bestTrackLink.SetIndex(pMCPoint->GetTrackID());
      const auto* pMCTrack = dynamic_cast<const CbmMCTrack*>(fpMCTracks->Get(bestTrackLink));
      LOG_IF(fatal, !pMCTrack) << fName << ": MC track object does not exist for hit " << iH << " and link: ";

      double t0MC = fpMCEventList->GetEventTime(bestPointLink);
      LOG_IF(fatal, t0MC < 0) << fName << ": MC time zero is lower then 0 ns: " << t0MC;

      // cut on the mc track quality
      if (!IsTrackSelected(pMCTrack, pMCPoint)) { continue; }

      {  // skip the case when the mc point participates to several hits
        int iE = fpMCEventList->GetEventIndex(bestTrackLink);
        if (vNofHitsPerMcTrack[iE][iSt][pMCPoint->GetTrackID()] != 1) { continue; }
      }

      // ----- MC point properties
      //
      double mass = pMCTrack->GetMass();
      //double charge = pMCTrack->GetCharge();
      //double pdg    = pMCTrack->GetPdgCode();

      // Entrance position and time
      double xMC = pMCPoint->GetXIn();
      double yMC = pMCPoint->GetYIn();
      double zMC = pMCPoint->GetZIn();
      double tMC = pMCPoint->GetTime() + t0MC;

      // MC point entrance momenta
      double pxMC = pMCPoint->GetPx();
      double pyMC = pMCPoint->GetPy();
      double pzMC = pMCPoint->GetPz();
      double pMC  = sqrt(pxMC * pxMC + pyMC * pyMC + pzMC * pzMC);

      // MC point exit momenta
      // double pxMCo = pMCPoint->GetPxOut();
      // double pyMCo = pMCPoint->GetPyOut();
      // double pzMCo = pMCPoint->GetPzOut();
      // double pMCo  = sqrt(pxMCo * pxMCo + pyMCo * pyMCo + pzMCo * pzMCo);

      // Position and time shifted to z-coordinate of the hit
      double shiftZ = zHit - zMC;  // Difference between hit and point z positions
      double xMCs   = xMC + shiftZ * pxMC / pzMC;
      double yMCs   = yMC + shiftZ * pyMC / pzMC;
      double uMCs   = xMCs * cosU + yMCs * sinU;
      double vMCs   = xMCs * cosV + yMCs * sinV;
      double tMCs   = tMC + shiftZ / (pzMC * phys::kSpeedOfLight) * sqrt(mass * mass + pMC * pMC);

      // Residuals
      double xRes = xHit - xMCs;
      double yRes = yHit - yMCs;
      double uRes = uHit - uMCs;
      double vRes = vHit - vMCs;
      double tRes = tHit - tMCs;
      double zRes = zHit - 0.5 * (pMCPoint->GetZIn() + pMCPoint->GetZOut());

      fvph_point_hit_delta_z[iSt]->Fill(zRes);

      fvph_res_x[iSt]->Fill(xRes);
      fvph_res_y[iSt]->Fill(yRes);
      fvph_res_u[iSt]->Fill(uRes);
      fvph_res_v[iSt]->Fill(vRes);
      fvph_res_t[iSt]->Fill(tRes);

      fvph_pull_x[iSt]->Fill(xRes / dxHit);
      fvph_pull_y[iSt]->Fill(yRes / dyHit);
      fvph_pull_u[iSt]->Fill(uRes / duHit);
      fvph_pull_v[iSt]->Fill(vRes / dvHit);
      fvph_pull_t[iSt]->Fill(tRes / dtHit);

      fvph_res_x_vs_x[iSt]->Fill(xHit, xRes);
      fvph_res_y_vs_y[iSt]->Fill(yHit, yRes);
      fvph_res_u_vs_u[iSt]->Fill(uHit, uRes);
      fvph_res_v_vs_v[iSt]->Fill(vHit, vRes);
      fvph_res_t_vs_t[iSt]->Fill(tHit, tRes);

      fvph_pull_x_vs_x[iSt]->Fill(xMCs, xRes / dxHit);
      fvph_pull_y_vs_y[iSt]->Fill(yMCs, yRes / dyHit);
      fvph_pull_u_vs_u[iSt]->Fill(uMCs, uRes / duHit);
      fvph_pull_v_vs_v[iSt]->Fill(vMCs, vRes / dvHit);
      fvph_pull_t_vs_t[iSt]->Fill(tMCs, tRes / dtHit);

      // fill summary histos

      fvph_point_hit_delta_z[nSt]->Fill(zRes);

      fvph_res_x[nSt]->Fill(xRes);
      fvph_res_y[nSt]->Fill(yRes);
      fvph_res_u[nSt]->Fill(uRes);
      fvph_res_v[nSt]->Fill(vRes);
      fvph_res_t[nSt]->Fill(tRes);

      fvph_pull_x[nSt]->Fill(xRes / dxHit);
      fvph_pull_y[nSt]->Fill(yRes / dyHit);
      fvph_pull_u[nSt]->Fill(uRes / duHit);
      fvph_pull_v[nSt]->Fill(vRes / dvHit);
      fvph_pull_t[nSt]->Fill(tRes / dtHit);

      fvph_res_x_vs_x[nSt]->Fill(xMCs, xRes);
      fvph_res_y_vs_y[nSt]->Fill(yMCs, yRes);
      fvph_res_u_vs_u[nSt]->Fill(uMCs, uRes);
      fvph_res_v_vs_v[nSt]->Fill(vMCs, vRes);
      fvph_res_t_vs_t[nSt]->Fill(tMCs, tRes);

      fvph_pull_x_vs_x[nSt]->Fill(xMCs, xRes / dxHit);
      fvph_pull_y_vs_y[nSt]->Fill(yMCs, yRes / dyHit);
      fvph_pull_u_vs_u[nSt]->Fill(uMCs, uRes / duHit);
      fvph_pull_v_vs_v[nSt]->Fill(vMCs, vRes / dvHit);
      fvph_pull_t_vs_t[nSt]->Fill(tMCs, tRes / dtHit);

      {
        const auto* pCluster = dynamic_cast<const CbmStsCluster*>(fpClusters->At(pHit->GetFrontClusterId()));
        assert(pCluster);
        int ndigis = pCluster->GetNofDigis();
        if (ndigis > fkMaxDigisInClusterForPulls) { ndigis = 0; }
        fvph_pull_u_Ndig[ndigis]->Fill(uRes / duHit);
      }

      {
        const auto* pCluster = dynamic_cast<const CbmStsCluster*>(fpClusters->At(pHit->GetBackClusterId()));
        assert(pCluster);
        int ndigis = pCluster->GetNofDigis();
        if (ndigis > fkMaxDigisInClusterForPulls) { ndigis = 0; }
        fvph_pull_v_Ndig[ndigis]->Fill(vRes / dvHit);
      }
    }
  }  // loop over hits: end

  // Fill hit reconstruction efficiencies
  if (IsMCUsed()) {
    for (int iE = 0; iE < nMCevents; ++iE) {
      int iFile   = fpMCEventList->GetFileIdByIndex(iE);
      int iEvent  = fpMCEventList->GetEventIdByIndex(iE);
      int nPoints = fpMCPoints->Size(iFile, iEvent);
      int nTracks = fpMCTracks->Size(iFile, iEvent);

      // If efficiency for the track at the station is already evaluated
      std::vector<std::vector<bool>> vIsTrackProcessed(nSt);
      for (int iSt = 0; iSt < nSt; iSt++) {
        vIsTrackProcessed[iSt].resize(nTracks, 0);
      }

      for (int iP = 0; iP < nPoints; ++iP) {
        const auto* pMCPoint = dynamic_cast<const CbmStsPoint*>(fpMCPoints->Get(iFile, iEvent, iP));
        LOG_IF(fatal, !pMCPoint) << fName << ": MC point does not exist for iFile = " << iFile
                                 << ", iEvent = " << iEvent << ", iP = " << iP;

        int address = pMCPoint->GetDetectorID();
        int iSt     = fpDetInterface->GetTrackingStationIndex(address);
        LOG_IF(fatal, iSt < 0 || iSt >= nSt)
          << fName << ": MC point for FEI = " << iFile << ", " << iEvent << ", " << iP << " and address " << address
          << " has wrong station index: iSt = " << iSt;


        double xMC = pMCPoint->GetXIn();
        double yMC = pMCPoint->GetYIn();
        double zMC = pMCPoint->GetZIn();

        {
          fvph_point_xy[iSt]->Fill(xMC, yMC);
          fvph_point_zx[iSt]->Fill(zMC, xMC);
          fvph_point_zy[iSt]->Fill(zMC, yMC);

          fvph_point_xy[nSt]->Fill(xMC, yMC);
          fvph_point_zx[nSt]->Fill(zMC, xMC);
          fvph_point_zy[nSt]->Fill(zMC, yMC);
        }

        int iTr = pMCPoint->GetTrackID();

        LOG_IF(fatal, iTr >= nTracks) << fName << ": index of MC track is out of range (point id = " << iP
                                      << ", event id = " << iE << ", track id = " << iTr << ')';

        const auto* pMCTrack = dynamic_cast<const CbmMCTrack*>(fpMCTracks->Get(iFile, iEvent, iTr));

        LOG_IF(fatal, pMCTrack == nullptr) << fName << ": null MC track pointer for file id = " << iFile
                                           << ", event id = " << iEvent << ", track id = " << iTr;

        // check efficiency only once per mc track per station
        if (vIsTrackProcessed[iSt][iTr]) { continue; }

        vIsTrackProcessed[iSt][iTr] = true;

        // cut on the mc track quality
        if (!IsTrackSelected(pMCTrack, pMCPoint)) { continue; }

        // Conditions under which point is accounted as reconstructed: point
        bool ifTrackHasHits = (vNofHitsPerMcTrack[iE][iSt][iTr] > 0);

        fvpe_reco_eff_vs_xy[iSt]->Fill(xMC, yMC, ifTrackHasHits);
        fvpe_reco_eff_vs_xy[nSt]->Fill(xMC, yMC, ifTrackHasHits);

      }  // loop over MC-points: end
    }    // loop over MC-events: end
  }      // MC is used: end
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaSts::InitDataBranches()
{
  // STS tracking detector interface
  fpDetInterface = CbmStsTrackingInterface::Instance();

  LOG_IF(fatal, !fpDetInterface) << "\033[1;31m" << fName << ": tracking detector interface is undefined\033[0m";

  // FairRootManager
  auto* pFairRootManager = FairRootManager::Instance();
  LOG_IF(fatal, !pFairRootManager) << "\033[1;31m" << fName << ": FairRootManager instance is a null pointer\033[0m";

  // Time-slice
  fpTimeSlice = dynamic_cast<CbmTimeSlice*>(pFairRootManager->GetObject("TimeSlice."));
  LOG_IF(fatal, !fpTimeSlice) << "\033[1;31m" << fName << ": time-slice branch is not found\033[0m";

  // ----- Reconstructed data-branches initialization

  // Hits container
  fpHits = dynamic_cast<TClonesArray*>(pFairRootManager->GetObject("StsHit"));
  LOG_IF(fatal, !fpHits) << "\033[1;31m" << fName << ": container of reconstructed hits in STS is not found\033[0m";

  // Clusters container
  fpClusters = dynamic_cast<TClonesArray*>(pFairRootManager->GetObject("StsCluster"));
  LOG_IF(fatal, !fpClusters) << "\033[1;31m" << fName << ": container of hit clusters in STS is not found\033[0m";


  // ----- MC-information branches initialization
  if (IsMCUsed()) {
    // MC manager
    fpMCDataManager = dynamic_cast<CbmMCDataManager*>(pFairRootManager->GetObject("MCDataManager"));
    LOG_IF(fatal, !fpMCDataManager) << "\033[1;31m" << fName << ": MC data manager branch is not found\033[0m";

    // MC event list
    fpMCEventList = dynamic_cast<CbmMCEventList*>(pFairRootManager->GetObject("MCEventList."));
    LOG_IF(fatal, !fpMCEventList) << "\033[1;31m" << fName << ": MC event list branch is not found\033[0m";

    // MC tracks
    fpMCTracks = fpMCDataManager->InitBranch("MCTrack");
    LOG_IF(fatal, !fpMCTracks) << "\033[1;31m" << fName << ": MC track branch is not found\033[0m";

    // MC points
    fpMCPoints = fpMCDataManager->InitBranch("StsPoint");
    LOG_IF(fatal, !fpMCTracks) << "\033[1;31m" << fName << ": MC point branch is not found for STS\033[0m";

    // Hit matches
    fpHitMatches = dynamic_cast<TClonesArray*>(pFairRootManager->GetObject("StsHitMatch"));
    LOG_IF(fatal, !fpHitMatches) << "\033[1;31m]" << fName << ": hit match branch is not found for STS\033[0m";
  }


  // get hit distribution ranges from the geometry

  int nSt = fpDetInterface->GetNtrackingStations();

  frXmin.resize(nSt + 1, 0.);
  frXmax.resize(nSt + 1, 0.);

  frYmin.resize(nSt + 1, 0.);
  frYmax.resize(nSt + 1, 0.);

  frZmin.resize(nSt + 1, 0.);
  frZmax.resize(nSt + 1, 0.);

  for (int i = nSt; i >= 0; --i) {
    {  // read the ranges for station i
      int j = (i == nSt ? 0 : i);

      frXmin[i] = fpDetInterface->GetXmin(j);
      frXmax[i] = fpDetInterface->GetXmax(j);

      frYmin[i] = fpDetInterface->GetYmin(j);
      frYmax[i] = fpDetInterface->GetYmax(j);

      frZmin[i] = fpDetInterface->GetZmin(j);
      frZmax[i] = fpDetInterface->GetZmax(j);
    }

    // update overall ranges

    frXmin[nSt] = std::min(frXmin[nSt], frXmin[i]);
    frXmax[nSt] = std::max(frXmax[nSt], frXmax[i]);

    frYmin[nSt] = std::min(frYmin[nSt], frYmin[i]);
    frYmax[nSt] = std::max(frYmax[nSt], frYmax[i]);

    frZmin[nSt] = std::min(frZmin[nSt], frZmin[i]);
    frZmax[nSt] = std::max(frZmax[nSt], frZmax[i]);
  }

  // add 5% margins for a better view

  for (int i = 0; i <= nSt; ++i) {
    double dx = 0.05 * fabs(frXmax[i] - frXmin[i]);
    frXmin[i] -= dx;
    frXmax[i] += dx;

    double dy = 0.05 * fabs(frYmax[i] - frYmin[i]);
    frYmin[i] -= dy;
    frYmax[i] += dy;

    double dz = 0.05 * fabs(frZmax[i] - frZmin[i]);
    frZmin[i] -= dz;
    frZmax[i] += dz;
  }


  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
// TODO: rename the method to Finish or Finalise or MakeSummary, since it does more than just canvases
//
InitStatus CbmCaInputQaSts::InitCanvases()
{
  gStyle->SetOptFit(1);
  int nSt = fpDetInterface->GetNtrackingStations();

  // ********************
  // ** Drawing options

  // Contours
  constexpr auto contColor = kOrange + 7;
  constexpr auto contWidth = 2;  // Line width
  constexpr auto contStyle = 2;  // Line style
  constexpr auto contFill  = 0;  // Fill style

  // **********************
  // ** summary for all stations

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("occ_hit", "Hit Occupancy", 1600, 800);
    canv->Divide2D(3);
    canv->cd(1);
    fvph_hit_xy[nSt]->DrawCopy("colz", "");
    canv->cd(2);
    fvph_hit_zx[nSt]->DrawCopy("colz", "");
    canv->cd(3);
    fvph_hit_zy[nSt]->DrawCopy("colz", "");
  }

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("occ_point", "Point Occupancy", 1600, 800);
    canv->Divide2D(3);

    canv->cd(1);
    fvph_point_xy[nSt]->DrawCopy("colz", "");
    canv->cd(2);
    fvph_point_zx[nSt]->DrawCopy("colz", "");
    canv->cd(3);
    fvph_point_zy[nSt]->DrawCopy("colz", "");
  }

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("residual", "Hit Residuals", 1600, 800);
    canv->Divide2D(6);
    canv->cd(1);
    fvph_res_x[nSt]->DrawCopy("colz", "");
    canv->cd(2);
    fvph_res_y[nSt]->DrawCopy("colz", "");
    canv->cd(3);
    fvph_res_t[nSt]->DrawCopy("colz", "");
    canv->cd(4);
    fvph_res_u[nSt]->DrawCopy("colz", "");
    canv->cd(5);
    fvph_res_v[nSt]->DrawCopy("colz", "");
  }

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("pull", "Hit Pulls", 1600, 800);
    canv->Divide2D(6);
    canv->cd(1);
    fvph_pull_x[nSt]->DrawCopy("colz", "");
    canv->cd(2);
    fvph_pull_y[nSt]->DrawCopy("colz", "");
    canv->cd(3);
    fvph_pull_t[nSt]->DrawCopy("colz", "");
    canv->cd(4);
    fvph_pull_u[nSt]->DrawCopy("colz", "");
    canv->cd(5);
    fvph_pull_v[nSt]->DrawCopy("colz", "");
  }

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("eff", "Hit Reconstruction Efficiency", 1600, 800);
    canv->Divide2D(2);
    canv->cd(1);
    fvpe_reco_eff_vs_xy[nSt]->DrawCopy("colz", "");
    canv->cd(2);
    fvph_reco_eff[nSt]->DrawCopy("colz", "");
  }

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("err", "Hit Errors", 1600, 800);
    canv->Divide2D(6);
    canv->cd(1);
    fvph_hit_dx[nSt]->DrawCopy();
    canv->cd(2);
    fvph_hit_dy[nSt]->DrawCopy();
    canv->cd(3);
    fvph_hit_dt[nSt]->DrawCopy();
    canv->cd(4);
    fvph_hit_du[nSt]->DrawCopy();
    canv->cd(5);
    fvph_hit_dv[nSt]->DrawCopy();
    canv->cd(6);
    fvph_hit_kuv[nSt]->DrawCopy();
  }

  {
    auto* canv = MakeQaObject<CbmQaCanvas>("other", "Other histograms", 1600, 800);
    canv->Divide2D(3);
    canv->cd(1);
    fvph_n_points_per_hit[nSt]->DrawCopy("colz", "");
    canv->cd(2);
    fvph_hit_station_delta_z[nSt]->DrawCopy("colz", "");
    canv->cd(3);
    fvph_point_hit_delta_z[nSt]->DrawCopy("colz", "");
  }


  // *********************************
  // ** Hit occupancies

  {  // ** Occupancy in XY plane
    auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/hit_xy", "", 1600, 800);
    canv->Divide2D(nSt);
    for (int iSt = 0; iSt < nSt; ++iSt) {
      canv->cd(iSt + 1);
      fvph_hit_xy[iSt]->DrawCopy("colz", "");

      // Square boarders of the active area of the station
      double stXmin = fpDetInterface->GetXmin(iSt);
      double stXmax = fpDetInterface->GetXmax(iSt);
      double stYmin = fpDetInterface->GetYmin(iSt);
      double stYmax = fpDetInterface->GetYmax(iSt);

      auto* pBox = new TBox(stXmin, stYmin, stXmax, stYmax);
      pBox->SetLineWidth(contWidth);
      pBox->SetLineStyle(contStyle);
      pBox->SetLineColor(contColor);
      pBox->SetFillStyle(contFill);
      pBox->Draw("SAME");
    }
  }

  {  // ** Occupancy in XZ plane
    auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/hit_zx", "", 1600, 800);
    canv->cd();
    fvph_hit_zx[nSt]->DrawCopy("colz", "");
    for (int iSt = 0; iSt < nSt; ++iSt) {
      // Station positions in detector IFS
      double stZmin = fpDetInterface->GetZmin(iSt);
      double stZmax = fpDetInterface->GetZmax(iSt);
      double stHmin = fpDetInterface->GetXmin(iSt);
      double stHmax = fpDetInterface->GetXmax(iSt);

      auto* pBox = new TBox(stZmin, stHmin, stZmax, stHmax);
      pBox->SetLineWidth(contWidth);
      pBox->SetLineStyle(contStyle);
      pBox->SetLineColor(contColor);
      pBox->SetFillStyle(contFill);
      pBox->Draw("SAME");
    }
  }

  {  // ** Occupancy in YZ plane
    auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/hit_zy", "", 1600, 800);
    canv->cd();
    fvph_hit_zy[nSt]->DrawCopy("colz", "");
    for (int iSt = 0; iSt < nSt; ++iSt) {
      // Station positions in detector IFS
      double stZmin = fpDetInterface->GetZmin(iSt);
      double stZmax = fpDetInterface->GetZmax(iSt);
      double stHmin = fpDetInterface->GetYmin(iSt);
      double stHmax = fpDetInterface->GetYmax(iSt);

      auto* pBox = new TBox(stZmin, stHmin, stZmax, stHmax);
      pBox->SetLineWidth(contWidth);
      pBox->SetLineStyle(contStyle);
      pBox->SetLineColor(contColor);
      pBox->SetFillStyle(contFill);
      pBox->Draw("SAME");
    }
  }

  // ************
  // ************

  if (IsMCUsed()) {

    // **********************
    // ** Point occupancies

    {  // ** Occupancy in XY plane
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/point_xy", "", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_point_xy[iSt]->DrawCopy("colz", "");

        // Square boarders of the active area of the station
        double stXmin = fpDetInterface->GetXmin(iSt);
        double stXmax = fpDetInterface->GetXmax(iSt);
        double stYmin = fpDetInterface->GetYmin(iSt);
        double stYmax = fpDetInterface->GetYmax(iSt);

        auto* pBox = new TBox(stXmin, stYmin, stXmax, stYmax);
        pBox->SetLineWidth(contWidth);
        pBox->SetLineStyle(contStyle);
        pBox->SetLineColor(contColor);
        pBox->SetFillStyle(contFill);
        pBox->Draw("SAME");
      }
    }

    {  // ** Occupancy in XZ plane
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/point_zx", "", 1600, 800);
      canv->cd();
      fvph_point_zx[nSt]->DrawCopy("colz", "");
      for (int iSt = 0; iSt < nSt; ++iSt) {
        // Station positions in detector IFS
        double stZmin = fpDetInterface->GetZmin(iSt);
        double stZmax = fpDetInterface->GetZmax(iSt);
        double stHmin = fpDetInterface->GetXmin(iSt);
        double stHmax = fpDetInterface->GetXmax(iSt);

        auto* pBox = new TBox(stZmin, stHmin, stZmax, stHmax);
        pBox->SetLineWidth(contWidth);
        pBox->SetLineStyle(contStyle);
        pBox->SetLineColor(contColor);
        pBox->SetFillStyle(contFill);
        pBox->Draw("SAME");
      }
    }

    {  // ** Occupancy in YZ plane
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/point_zy", "", 1600, 800);
      canv->cd();
      fvph_point_zy[nSt]->DrawCopy("colz", "");
      for (int iSt = 0; iSt < nSt; ++iSt) {
        // Station positions in detector IFS
        double stZmin = fpDetInterface->GetZmin(iSt);
        double stZmax = fpDetInterface->GetZmax(iSt);
        double stHmin = fpDetInterface->GetYmin(iSt);
        double stHmax = fpDetInterface->GetYmax(iSt);

        auto* pBox = new TBox(stZmin, stHmin, stZmax, stHmax);
        pBox->SetLineWidth(contWidth);
        pBox->SetLineStyle(contStyle);
        pBox->SetLineColor(contColor);
        pBox->SetFillStyle(contFill);
        pBox->Draw("SAME");
      }
    }

    // **************
    // ** Residuals

    {  // x-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/res_x", "Residuals for x coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_res_x[iSt]->DrawCopy("", "");
      }
    }

    {  // y-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/res_y", "Residuals for y coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_res_y[iSt]->DrawCopy("", "");
      }
    }

    {  // u-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/res_u", "Residuals for u coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_res_u[iSt]->DrawCopy("", "");
      }
    }

    {  // v-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/res_v", "Residuals for v coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_res_v[iSt]->DrawCopy("", "");
      }
    }

    {  // t-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/res_t", "Residuals for time", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_res_t[iSt]->DrawCopy("", "");
      }
    }


    // **********
    // ** Pulls

    {  // x-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/pull_x", "Pulls for x coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_pull_x[iSt]->DrawCopy("", "");
      }
    }

    {  // y-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/pull_y", "Pulls for y coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_pull_y[iSt]->DrawCopy("", "");
      }
    }

    {  // u-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/pull_u", "Pulls for u coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_pull_u[iSt]->DrawCopy("", "");
      }
    }

    {  // v-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/pull_v", "Pulls for v coordinate", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_pull_v[iSt]->DrawCopy("", "");
      }
    }

    {  // t-coordinate
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/pull_t", "Pulls for time", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvph_pull_t[iSt]->DrawCopy("", "");
      }
    }

    {  // u-coordinate vs N digis
      auto* canv = MakeQaObject<CbmQaCanvas>("vs N digi/pull_u", "Pulls for u coordinate vs N digis", 1600, 800);
      canv->Divide2D(fkMaxDigisInClusterForPulls + 1);
      for (int i = 0; i <= fkMaxDigisInClusterForPulls; ++i) {
        canv->cd((i > 0) ? i : fkMaxDigisInClusterForPulls + 1);
        fvph_pull_u_Ndig[i]->DrawCopy("", "");
      }
    }

    {  // v-coordinate vs N digis
      auto* canv = MakeQaObject<CbmQaCanvas>("vs N digi/pull_v", "Pulls for v coordinate vs N digis", 1600, 800);
      canv->Divide2D(fkMaxDigisInClusterForPulls + 1);
      for (int i = 0; i <= fkMaxDigisInClusterForPulls; ++i) {
        canv->cd((i > 0) ? i : fkMaxDigisInClusterForPulls + 1);
        fvph_pull_v_Ndig[i]->DrawCopy("", "");
      }
    }

    // ************************************
    // ** Hit reconstruction efficiencies

    {
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/reco_eff", "Hit efficiencies in xy bins", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        //fvph_reco_eff[iSt]->SetMarkerStyle(20);
        fvph_reco_eff[iSt]->DrawCopy();
      }
    }

    {
      auto* canv = MakeQaObject<CbmQaCanvas>("vs Station/reco_eff_vs_xy", "Hit efficiencies wrt x and y", 1600, 800);
      canv->Divide2D(nSt);
      for (int iSt = 0; iSt < nSt; ++iSt) {
        canv->cd(iSt + 1);
        fvpe_reco_eff_vs_xy[iSt]->DrawCopy("colz", "");
      }
    }
  }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaSts::InitHistograms()
{
  int nSt = fpDetInterface->GetNtrackingStations();

  //SetStoringMode(EStoringMode::kSAMEDIR);
  SetStoringMode(EStoringMode::kSUBDIR);

  // ----- Subdirectories
  MakeQaDirectory("Summary");
  MakeQaDirectory("Summary/vs Station");
  MakeQaDirectory("Summary/vs N digi");
  MakeQaDirectory("All stations");

  // ----- Histograms initialization
  fvph_hit_xy.resize(nSt + 1, nullptr);
  fvph_hit_zy.resize(nSt + 1, nullptr);
  fvph_hit_zx.resize(nSt + 1, nullptr);

  fvph_hit_station_delta_z.resize(nSt + 1, nullptr);

  fvph_hit_dx.resize(nSt + 1, nullptr);
  fvph_hit_dy.resize(nSt + 1, nullptr);
  fvph_hit_dt.resize(nSt + 1, nullptr);
  fvph_hit_dv.resize(nSt + 1, nullptr);
  fvph_hit_kuv.resize(nSt + 1, nullptr);
  fvph_hit_du.resize(nSt + 1, nullptr);

  for (int iSt = 0; iSt <= nSt; ++iSt) {
    TString sF = "";  // Subfolder
    sF += (iSt == nSt) ? "All stations/" : Form("Station %d/", iSt);
    TString nsuff = (iSt == nSt) ? "" : Form("_st%d", iSt);               // Histogram name suffix
    TString tsuff = (iSt == nSt) ? "" : Form(" in STS station %d", iSt);  // Histogram title suffix
    TString sN    = "";
    TString sT    = "";

    // place directories in a prefered order
    MakeQaDirectory(sF + "occup/");
    if (IsMCUsed()) {
      MakeQaDirectory(sF + "res/");
      MakeQaDirectory(sF + "pull/");
      MakeQaDirectory(sF + "eff/");
    }
    MakeQaDirectory(sF + "err/");

    int nBinsZ = (iSt < nSt) ? kNbins : kNbinsZ;

    // Hit occupancy
    sN = (TString) "hit_xy" + nsuff;
    sT = (TString) "Hit occupancy in xy-Plane" + tsuff + ";x_{hit} [cm];y_{hit} [cm]";
    fvph_hit_xy[iSt] =
      MakeQaObject<TH2F>(sF + "occup/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, frYmin[iSt], frYmax[iSt]);

    sN = (TString) "hit_zx" + nsuff;
    sT = (TString) "Hit occupancy in xz-Plane" + tsuff + ";z_{hit} [cm];x_{hit} [cm]";
    fvph_hit_zx[iSt] =
      MakeQaObject<TH2F>(sF + "occup/" + sN, sT, nBinsZ, frZmin[iSt], frZmax[iSt], kNbins, frXmin[iSt], frXmax[iSt]);

    sN = (TString) "hit_zy" + nsuff;
    sT = (TString) "Hit occupancy in yz-plane" + tsuff + ";z_{hit} [cm];y_{hit} [cm]";
    fvph_hit_zy[iSt] =
      MakeQaObject<TH2F>(sF + "occup/" + sN, sT, nBinsZ, frZmin[iSt], frZmax[iSt], kNbins, frYmin[iSt], frYmax[iSt]);

    // Hit errors
    sN               = (TString) "hit_dx" + nsuff;
    sT               = (TString) "Hit position error along x-axis" + tsuff + ";dx_{hit} [cm]";
    fvph_hit_dx[iSt] = MakeQaObject<TH1F>(sF + "err/" + sN, sT, kNbins, kRHitDx[0], kRHitDx[1]);

    sN               = (TString) "hit_dy" + nsuff;
    sT               = (TString) "Hit position error along y-axis" + tsuff + ";dy_{hit} [cm]";
    fvph_hit_dy[iSt] = MakeQaObject<TH1F>(sF + "err/" + sN, sT, kNbins, kRHitDy[0], kRHitDy[1]);

    sN               = (TString) "hit_du" + nsuff;
    sT               = (TString) "Hit position error across front strips" + tsuff + ";du_{hit} [cm]";
    fvph_hit_du[iSt] = MakeQaObject<TH1F>(sF + "err/" + sN, sT, kNbins, kRHitDu[0], kRHitDu[1]);

    sN               = (TString) "hit_dv" + nsuff;
    sT               = (TString) "Hit position error across back strips" + tsuff + ";dv_{hit} [cm]";
    fvph_hit_dv[iSt] = MakeQaObject<TH1F>(sF + "err/" + sN, sT, kNbins, kRHitDv[0], kRHitDv[1]);

    sN = (TString) "hit_kuv" + nsuff;
    sT = (TString) "Hit error correlation between front and back strips" + tsuff + ";kuv_{hit} [unitless]";
    fvph_hit_kuv[iSt] = MakeQaObject<TH1F>(sF + "err/" + sN, sT, kNbins / 2 * 2 + 1, -1.1, 1.1);

    sN               = (TString) "hit_dt" + nsuff;
    sT               = (TString) "Hit time error" + tsuff + ";dt_{hit} [ns]";
    fvph_hit_dt[iSt] = MakeQaObject<TH1F>(sF + "err/" + sN, sT, kNbins, kRHitDt[0], kRHitDt[1]);

    sN = (TString) "hit_station_delta_z" + nsuff;
    sT = (TString) "Different between hit and station z-positions" + tsuff + ";z_{hit} - z_{st} [cm]";
    fvph_hit_station_delta_z[iSt] = MakeQaObject<TH1F>(sF + sN, sT, kNbins, -5., 5.);

  }  // loop over station index: end

  // ----- Initialize histograms, which are use MC-information
  if (IsMCUsed()) {
    // Resize histogram vectors
    fvph_n_points_per_hit.resize(nSt + 1, nullptr);
    fvph_point_xy.resize(nSt + 1, nullptr);
    fvph_point_zx.resize(nSt + 1, nullptr);
    fvph_point_zy.resize(nSt + 1, nullptr);
    fvph_point_hit_delta_z.resize(nSt + 1, nullptr);
    fvph_res_x.resize(nSt + 1, nullptr);
    fvph_res_y.resize(nSt + 1, nullptr);
    fvph_res_u.resize(nSt + 1, nullptr);
    fvph_res_v.resize(nSt + 1, nullptr);
    fvph_res_t.resize(nSt + 1, nullptr);
    fvph_pull_x.resize(nSt + 1, nullptr);
    fvph_pull_y.resize(nSt + 1, nullptr);
    fvph_pull_u.resize(nSt + 1, nullptr);

    fvph_pull_u_Ndig.resize(fkMaxDigisInClusterForPulls + 1, nullptr);
    fvph_pull_v_Ndig.resize(fkMaxDigisInClusterForPulls + 1, nullptr);

    fvph_pull_v.resize(nSt + 1, nullptr);
    fvph_pull_t.resize(nSt + 1, nullptr);
    fvph_res_x_vs_x.resize(nSt + 1, nullptr);
    fvph_res_y_vs_y.resize(nSt + 1, nullptr);
    fvph_res_u_vs_u.resize(nSt + 1, nullptr);
    fvph_res_v_vs_v.resize(nSt + 1, nullptr);
    fvph_res_t_vs_t.resize(nSt + 1, nullptr);
    fvph_pull_x_vs_x.resize(nSt + 1, nullptr);
    fvph_pull_y_vs_y.resize(nSt + 1, nullptr);
    fvph_pull_u_vs_u.resize(nSt + 1, nullptr);
    fvph_pull_v_vs_v.resize(nSt + 1, nullptr);
    fvph_pull_t_vs_t.resize(nSt + 1, nullptr);
    fvpe_reco_eff_vs_xy.resize(nSt + 1, nullptr);
    fvph_reco_eff.resize(nSt + 1, nullptr);

    for (int iSt = 0; iSt <= nSt; ++iSt) {
      TString sF = "";  // Subfolder
      sF += (iSt == nSt) ? "All stations/" : Form("Station %d/", iSt);
      TString nsuff = (iSt == nSt) ? "" : Form("_st%d", iSt);               // Histogram name suffix
      TString tsuff = (iSt == nSt) ? "" : Form(" in STS station %d", iSt);  // Histogram title suffix
      TString sN    = "";
      TString sT    = "";

      sN                         = (TString) "n_points_per_hit" + nsuff;
      sT                         = (TString) "Number of points per hit" + tsuff + ";N_{point}/hit";
      fvph_n_points_per_hit[iSt] = MakeQaObject<TH1F>(sF + sN, sT, 10, -0.5, 9.5);

      // Point occupancy
      sN = (TString) "point_xy" + nsuff;
      sT = (TString) "Point occupancy in XY plane" + tsuff + ";x_{MC} [cm];y_{MC} [cm]";
      fvph_point_xy[iSt] =
        MakeQaObject<TH2F>(sF + "occup/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, frYmin[iSt], frYmax[iSt]);

      sN = (TString) "point_zx" + nsuff;
      sT = (TString) "Point Occupancy in XZ plane" + tsuff + ";z_{MC} [cm];x_{MC} [cm]";
      fvph_point_zx[iSt] =
        MakeQaObject<TH2F>(sF + "occup/" + sN, sT, kNbinsZ, frZmin[iSt], frZmax[iSt], kNbins, frXmin[iSt], frXmax[iSt]);

      sN = (TString) "point_zy" + nsuff;
      sT = (TString) "Point Occupancy in YZ Plane" + tsuff + ";z_{MC} [cm];y_{MC} [cm]";
      fvph_point_zy[iSt] =
        MakeQaObject<TH2F>(sF + "occup/" + sN, sT, kNbinsZ, frZmin[iSt], frZmax[iSt], kNbins, frYmin[iSt], frYmax[iSt]);

      // Difference between MC input z and hit z coordinates
      sN = (TString) "point_hit_delta_z" + nsuff;
      sT = (TString) "Distance between STS point and hit along z axis" + tsuff + ";z_{reco} - z_{MC} [cm]";
      fvph_point_hit_delta_z[iSt] = MakeQaObject<TH1F>(sF + sN, sT, kNbins, -0.05, 0.05);

      sN              = (TString) "res_x" + nsuff;
      sT              = (TString) "Residuals for X" + tsuff + ";x_{reco} - x_{MC} [cm]";
      fvph_res_x[iSt] = MakeQaObject<TH1F>(sF + "res/" + sN, sT, kNbins, kRResX[0], kRResX[1]);

      sN              = (TString) "res_y" + nsuff;
      sT              = (TString) "Residuals for Y" + tsuff + ";y_{reco} - y_{MC} [cm]";
      fvph_res_y[iSt] = MakeQaObject<TH1F>(sF + "res/" + sN, sT, kNbins, kRResY[0], kRResY[1]);

      sN              = (TString) "res_u" + nsuff;
      sT              = (TString) "Residuals for Front strip coordinate" + tsuff + ";u_{reco} - u_{MC} [cm]";
      fvph_res_u[iSt] = MakeQaObject<TH1F>(sF + "res/" + sN, sT, kNbins, kRResU[0], kRResU[1]);

      sN              = (TString) "res_v" + nsuff;
      sT              = (TString) "Residuals for Back strip coordinate" + tsuff + ";v_{reco} - v_{MC} [cm]";
      fvph_res_v[iSt] = MakeQaObject<TH1F>(sF + "res/" + sN, sT, kNbins, kRResV[0], kRResV[1]);

      sN              = (TString) "res_t" + nsuff;
      sT              = (TString) "Residuals for Time" + tsuff + ";t_{reco} - t_{MC} [ns]";
      fvph_res_t[iSt] = MakeQaObject<TH1F>(sF + "res/" + sN, sT, kNbins, kRResT[0], kRResT[1]);

      sN               = (TString) "pull_x" + nsuff;
      sT               = (TString) "Pulls for X" + tsuff + ";(x_{reco} - x_{MC}) / #sigma_{x}^{reco}";
      fvph_pull_x[iSt] = MakeQaObject<TH1F>(sF + "pull/" + sN, sT, kNbins, kRPull[0], kRPull[1]);

      sN               = (TString) "pull_y" + nsuff;
      sT               = (TString) "Pulls for Y" + tsuff + ";(y_{reco} - y_{MC}) / #sigma_{y}^{reco}";
      fvph_pull_y[iSt] = MakeQaObject<TH1F>(sF + "pull/" + sN, sT, kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "pull_u" + nsuff;
      sT = (TString) "Pulls for Front strip coordinate" + tsuff + ";(u_{reco} - u_{MC}) / #sigma_{u}^{reco}";
      fvph_pull_u[iSt] = MakeQaObject<TH1F>(sF + "pull/" + sN, sT, kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "pull_v" + nsuff;
      sT = (TString) "Pulls for Back strip coordinate" + tsuff + ";(v_{reco} - v_{MC}) / #sigma_{v}^{reco}";
      fvph_pull_v[iSt] = MakeQaObject<TH1F>(sF + "pull/" + sN, sT, kNbins, kRPull[0], kRPull[1]);

      sN               = (TString) "pull_t" + nsuff;
      sT               = (TString) "Pulls for Time" + tsuff + ";(t_{reco} - t_{MC}) / #sigma_{t}^{reco}";
      fvph_pull_t[iSt] = MakeQaObject<TH1F>(sF + "pull/" + sN, sT, kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "res_x_vs_x" + nsuff;
      sT = (TString) "Residuals for X" + tsuff + ";x_{reco} [cm];x_{reco} - x_{MC} [cm]";
      fvph_res_x_vs_x[iSt] =
        MakeQaObject<TH2F>(sF + "res/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, kRResX[0], kRResX[1]);

      sN = (TString) "res_y_vs_y" + nsuff;
      sT = (TString) "Residuals for Y" + tsuff + ";y_{reco} [cm];y_{reco} - y_{MC} [cm]";
      fvph_res_y_vs_y[iSt] =
        MakeQaObject<TH2F>(sF + "res/" + sN, sT, kNbins, frYmin[iSt], frYmax[iSt], kNbins, kRResY[0], kRResY[1]);

      sN = (TString) "res_u_vs_u" + nsuff;
      sT = (TString) "Residuals for Front strip coordinate" + tsuff + ";u_{reco} [cm];u_{reco} - u_{MC} [cm]";
      fvph_res_u_vs_u[iSt] =
        MakeQaObject<TH2F>(sF + "res/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, kRResU[0], kRResU[1]);

      sN = (TString) "res_v_vs_v" + nsuff;
      sT = (TString) "Residuals for Back strip coordinate" + tsuff + ";v_{reco} [cm];v_{reco} - v_{MC} [cm]";
      fvph_res_v_vs_v[iSt] =
        MakeQaObject<TH2F>(sF + "res/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, kRResV[0], kRResV[1]);

      sN                   = (TString) "res_t_vs_t" + nsuff;
      sT                   = (TString) "Residuals for Time" + tsuff + "t_{reco} [ns];t_{reco} - t_{MC} [ns]";
      fvph_res_t_vs_t[iSt] = MakeQaObject<TH2F>(sF + "res/" + sN, sT, kNbins, 0, 0, kNbins, kRResT[0], kRResT[1]);

      sN = (TString) "pull_x_vs_x" + nsuff;
      sT = (TString) "Pulls for X" + tsuff + "x_{reco} [cm];(x_{reco} - x_{MC}) / #sigma_{x}^{reco}";
      fvph_pull_x_vs_x[iSt] =
        MakeQaObject<TH2F>(sF + "pull/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "pull_y_vs_y" + nsuff;
      sT = (TString) "Pulls for Y" + tsuff + ";y_{reco} [cm];(y_{reco} - y_{MC}) / #sigma_{y}^{reco}";
      fvph_pull_y_vs_y[iSt] =
        MakeQaObject<TH2F>(sF + "pull/" + sN, sT, kNbins, frYmin[iSt], frYmax[iSt], kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "pull_u_vs_u" + nsuff;
      sT =
        (TString) "Pulls for Front strip coordinate" + tsuff + ";u_{reco} [cm];(u_{reco} - u_{MC}) / #sigma_{u}^{reco}";
      fvph_pull_u_vs_u[iSt] =
        MakeQaObject<TH2F>(sF + "pull/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "pull_v_vs_v" + nsuff;
      sT =
        (TString) "Pulls for Back strip coordinate" + tsuff + ";v_{reco} [cm];(v_{reco} - v_{MC}) / #sigma_{v}^{reco}";
      fvph_pull_v_vs_v[iSt] =
        MakeQaObject<TH2F>(sF + "pull/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt], kNbins, kRPull[0], kRPull[1]);

      sN = (TString) "pull_t_vs_t" + nsuff;
      sT = (TString) "Pulls for Time" + tsuff + ";t_{reco} [ns];(t_{reco} - t_{MC}) / #sigma_{t}^{reco}";
      fvph_pull_t_vs_t[iSt] = MakeQaObject<TH2F>(sF + "pull/" + sN, sT, kNbins, 0, 0, kNbins, kRPull[0], kRPull[1]);

      sN                       = (TString) "reco_eff_vs_xy" + nsuff;
      sT                       = (TString) "Hit rec. efficiency in XY" + tsuff + ";x_{MC} [cm];y_{MC} [cm];#epsilon";
      fvpe_reco_eff_vs_xy[iSt] = MakeQaObject<TProfile2D>(sF + "eff/" + sN, sT, kNbins, frXmin[iSt], frXmax[iSt],
                                                          kNbins, frYmin[iSt], frYmax[iSt]);

      sN                 = (TString) "reco_eff" + nsuff;
      sT                 = (TString) "Hit rec. efficiency in XY bins" + tsuff + ";eff";
      fvph_reco_eff[iSt] = MakeQaObject<TH1F>(sF + "eff/" + sN, sT, 130, -0.005, 1.30 - 0.005);
    }  // stations

    for (int idig = 0; idig <= fkMaxDigisInClusterForPulls; idig++) {
      TString sN = "All stations/pull/Ndigi/pull_u_";
      TString sT = "Pulls for Front strip coordinate, hits with ";
      if (idig == 0) {
        sN += Form("%d+_digi", fkMaxDigisInClusterForPulls + 1);
        sT += Form(">=%d digi", fkMaxDigisInClusterForPulls + 1);
      }
      else {
        sN += Form("%d_digi", idig);
        sT += Form("%d digi", idig);
      }
      sT += "; (u_{reco} - u_{MC}) / #sigma_{u}^{reco}";
      fvph_pull_u_Ndig[idig] = MakeQaObject<TH1F>(sN, sT, kNbins, kRPull[0], kRPull[1]);
    }

    for (int idig = 0; idig <= fkMaxDigisInClusterForPulls; idig++) {
      TString sN = "All stations/pull/Ndigi/pull_v_";
      TString sT = "Pulls for Back strip coordinate, hits with ";
      if (idig == 0) {
        sN += Form("%d+_digi", fkMaxDigisInClusterForPulls + 1);
        sT += Form(">=%d digi", fkMaxDigisInClusterForPulls + 1);
      }
      else {
        sN += Form("%d_digi", idig);
        sT += Form("%d digi", idig);
      }
      sT += "; (v_{reco} - v_{MC}) / #sigma_{v}^{reco}";
      fvph_pull_v_Ndig[idig] = MakeQaObject<TH1F>(sN, sT, kNbins, kRPull[0], kRPull[1]);
    }
  }
  return kSUCCESS;
}


// ---------------------------------------------------------------------------------------------------------------------
//
Bool_t CbmCaInputQaSts::IsTrackSelected(const CbmMCTrack* track, const CbmStsPoint* point) const
{

  if (kTrackSelectionPrimary && track->GetMotherId() >= 0) { return false; }

  double px = point->GetPxOut();
  double py = point->GetPyOut();
  double pz = point->GetPzOut();
  double p  = sqrt(px * px + py * py + pz * pz);

  if (pz <= 0.) { return false; }

  if (p < kTrackSelectionMomentum) { return false; }

  if (TMath::ATan2(sqrt(px * px + py * py), pz) * TMath::RadToDeg() > kTrackSelectionAngle) { return false; }

  return true;
}
