/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmCaInputQaSts.h
/// \date   13.01.2023
/// \brief  QA-task for CA tracking input from MuCh detector (header)
/// \author S.Zharko <s.zharko@gsi.de>


#ifndef CbmCaInputQaSts_h
#define CbmCaInputQaSts_h 1

#include "CbmCaInputQaBase.h"
#include "CbmMCDataManager.h"
#include "CbmQaTask.h"

#include "TMath.h"

#include <set>
#include <unordered_map>
#include <vector>

class CbmMatch;
class CbmMCEventList;
class CbmMCDataArray;
class CbmMCDataManager;
class CbmMCTrack;
class CbmStsHit;
class CbmStsPoint;
class CbmStsTrackingInterface;
class CbmTimeSlice;
class TClonesArray;
class TH1F;
class TH2F;
class TProfile;
class TProfile2D;

/// A QA-task class, which provides assurance of MuCh hits and geometry
class CbmCaInputQaSts : public CbmQaTask, public CbmCaInputQaBase {
public:
  /// Constructor from parameters
  /// \param  verbose   Verbose level
  /// \param  isMCUsed  Flag, whether MC information is available for this task
  CbmCaInputQaSts(int verbose, bool isMCUsed);

protected:
  // ********************************************
  // ** Virtual method override from CbmQaTask **
  // ********************************************

  /// Checks results of the QA and returns some flag
  bool Check();

  /// Initializes data branches
  InitStatus InitDataBranches();

  /// Initializes canvases
  InitStatus InitCanvases();

  /// Initializes histograms
  InitStatus InitHistograms();

  /// Fills histograms per event or time-slice
  void FillHistograms();

  /// De-initializes histograms
  void DeInit();

  /// Checks ranges for mean and standard deviation
  /// \return  False, if variable exits the range
  bool CheckRangePull(TH1* h)
  {
    return CbmQaTask::CheckRange(h, fPullMeanThrsh, 1. - fPullWidthThrsh, 1. + fPullWidthThrsh);
  }

  /// \brief checks if the track at mc point passes the track selection cuts
  /// \return true when selected
  Bool_t IsTrackSelected(const CbmMCTrack* track, const CbmStsPoint* point) const;

private:
  // ----- Data branches
  CbmStsTrackingInterface* fpDetInterface = nullptr;  ///< Instance of detector interface

  CbmTimeSlice* fpTimeSlice = nullptr;  ///< Pointer to current time-slice

  TClonesArray* fpHits     = nullptr;  ///< Array of hits
  TClonesArray* fpClusters = nullptr;  ///< Array of hit clusters

  CbmMCDataManager* fpMCDataManager = nullptr;  ///< MC data manager
  CbmMCEventList* fpMCEventList     = nullptr;  ///< MC event list
  CbmMCDataArray* fpMCTracks        = nullptr;  ///< Array of MC tracks
  CbmMCDataArray* fpMCPoints        = nullptr;  ///< Array of MC points

  TClonesArray* fpHitMatches = nullptr;  ///< Array of hit matches

  // ----- Track selection for efficiencies and pulls

  static constexpr bool kTrackSelectionPrimary {true};    // must the track come from the primary vertex
  static constexpr double kTrackSelectionMomentum {0.1};  // track momentum GeV when it exits the tracking station
  static constexpr double kTrackSelectionAngle {60};      // track incl. angle [grad] when it exits the tracking station

  // ----- Histograms

  static constexpr int kNbins  = 200;  ///< General number of bins
  static constexpr int kNbinsZ = 800;  ///< Number of bins for z axis in overall views

  static constexpr double kRHitDx[2] = {0., .0050};  ///< Range for hit x coordinate error [cm]
  static constexpr double kRHitDy[2] = {0., .0200};  ///< Range for hit y coordinate error [cm]
  static constexpr double kRHitDu[2] = {0., .0050};  ///< Range for hit u coordinate error [cm]
  static constexpr double kRHitDv[2] = {0., .0050};  ///< Range for hit v coordinate error [cm]
  static constexpr double kRHitDt[2] = {0., 10.};    ///< Range for hit time error [ns]

  static constexpr double kRResX[2] = {-.02, .02};  ///< Range for residual of x coordinate [cm]
  static constexpr double kRResY[2] = {-.1, .1};    ///< Range for residual of y coordinate [cm]
  static constexpr double kRResU[2] = {-.02, .02};  ///< Range for residual of u coordinate [cm]
  static constexpr double kRResV[2] = {-.02, .02};  ///< Range for residual of v coordinate [cm]
  static constexpr double kRResT[2] = {-25., 25.};  ///< Range for residual of time [ns]

  static constexpr double kRPull[2] = {-10., 10.};  ///< Range for pull histograms coordinate

  std::vector<double> frXmin;  ///< Range for hit x coordinate [cm]
  std::vector<double> frXmax;  ///< Range for hit x coordinate [cm]

  std::vector<double> frYmin;  ///< Range for hit y coordinate [cm]
  std::vector<double> frYmax;  ///< Range for hit y coordinate [cm]

  std::vector<double> frZmin;  ///< Range for hit z coordinate [cm]
  std::vector<double> frZmax;  ///< Range for hit z coordinate [cm]

  // NOTE: the last element of each vector stands for integral distribution over all stations

  // Hit occupancy
  std::vector<TH2F*> fvph_hit_xy;  ///< hit y vs x in different stations
  std::vector<TH2F*> fvph_hit_zx;  ///< hit x vs z in different stations
  std::vector<TH2F*> fvph_hit_zy;  ///< hit y vs z in different stations

  std::vector<TH1F*> fvph_hit_station_delta_z;  ///< Difference between station and hit z positions [cm]

  // Hit errors
  std::vector<TH1F*> fvph_hit_dx;
  std::vector<TH1F*> fvph_hit_dy;
  std::vector<TH1F*> fvph_hit_du;
  std::vector<TH1F*> fvph_hit_dv;
  std::vector<TH1F*> fvph_hit_kuv;
  std::vector<TH1F*> fvph_hit_dt;

  // MC points occupancy
  std::vector<TH1F*> fvph_n_points_per_hit;  ///< number of points per one hit in different stations

  std::vector<TH2F*> fvph_point_xy;  ///< point y vs x in different stations
  std::vector<TH2F*> fvph_point_zx;  ///< point x vs z in different stations
  std::vector<TH2F*> fvph_point_zy;  ///< point y vs z in different stations

  std::vector<TH1F*> fvph_point_hit_delta_z;  ///< difference between z of hit and MC point in different stations

  // Residuals
  std::vector<TH1F*> fvph_res_x;  ///< residual for x coordinate in different stations
  std::vector<TH1F*> fvph_res_y;  ///< residual for y coordinate in different stations
  std::vector<TH1F*> fvph_res_u;  ///< residual for u coordinate in different stations
  std::vector<TH1F*> fvph_res_v;  ///< residual for v coordinate in different stations
  std::vector<TH1F*> fvph_res_t;  ///< residual for t coordinate in different stations

  std::vector<TH2F*> fvph_res_x_vs_x;  ///< residual for x coordinate in different stations
  std::vector<TH2F*> fvph_res_y_vs_y;  ///< residual for y coordinate in different stations
  std::vector<TH2F*> fvph_res_u_vs_u;  ///< residual for u coordinate in different stations
  std::vector<TH2F*> fvph_res_v_vs_v;  ///< residual for v coordinate in different stations
  std::vector<TH2F*> fvph_res_t_vs_t;  ///< residual for t coordinate in different stations

  // Pulls
  std::vector<TH1F*> fvph_pull_x;  ///< pull for x coordinate in different stations
  std::vector<TH1F*> fvph_pull_y;  ///< pull for y coordinate in different stations
  std::vector<TH1F*> fvph_pull_u;  ///< pull for u coordinate in different stations
  std::vector<TH1F*> fvph_pull_v;  ///< pull for v coordinate in different stations
  std::vector<TH1F*> fvph_pull_t;  ///< pull for t coordinate in different stations

  const int fkMaxDigisInClusterForPulls {5};  ///< max digis in cluster for separate histogramming of puls

  std::vector<TH1F*> fvph_pull_u_Ndig;  ///< pull for u coordinate, depending on N digis in the cluster
  std::vector<TH1F*> fvph_pull_v_Ndig;  ///< pull for v coordinate, depending on N digis in the cluster

  std::vector<TH2F*> fvph_pull_x_vs_x;  ///< pull for x coordinate in different stations
  std::vector<TH2F*> fvph_pull_y_vs_y;  ///< pull for y coordinate in different stations
  std::vector<TH2F*> fvph_pull_u_vs_u;  ///< pull for u coordinate in different stations
  std::vector<TH2F*> fvph_pull_v_vs_v;  ///< pull for v coordinate in different stations
  std::vector<TH2F*> fvph_pull_t_vs_t;  ///< pull for t coordinate in different stations

  // Hit efficiencies
  std::vector<TProfile2D*> fvpe_reco_eff_vs_xy;  ///< Efficiency of hit reconstruction vs. x and y coordinates of MC
  std::vector<TH1F*> fvph_reco_eff;  ///< Distribution of hit reconstruction efficiency in bins of fvpe_reco_eff_vs_xy

  ClassDef(CbmCaInputQaSts, 0);
};

#endif  // CbmCaInputQaSts_h
